# -*- coding: utf-8 -*-
from accelbrainbase.controllable_model import ControllableModel

from accelbrainbase.iteratabledata._torch.ssda_iterator import SSDAIterator
from accelbrainbase.regularizatable_data import RegularizatableData
from accelbrainbase.computableloss._torch.ssda_loss import SSDALoss
from accelbrainbase.observabledata._torch.neural_networks import NeuralNetworks
from accelbrainbase.observabledata._torch.convolutional_neural_networks import ConvolutionalNeuralNetworks
import numpy as np
from logging import getLogger
import torch
from torch import nn


class SSDAController(nn.Module, ControllableModel):
    '''
    Self-supervised domain adaptation.

    References:
        - Jing, L., & Tian, Y. (2020). Self-supervised visual feature learning with deep neural networks: A survey. IEEE Transactions on Pattern Analysis and Machine Intelligence.
        - Xu, J., Xiao, L., & López, A. M. (2019). Self-supervised domain adaptation for computer vision tasks. IEEE Access, 7, 156694-156706., p156698.

    '''
    
    def __init__(
        self,
        encoder,
        classifier,
        pretext_task_model,
        ssda_loss,
        learning_rate=1e-05,
        ctx="cpu",
        scale=1.0,
        tol=3.0,
        est=1e-08,
        wd=0.0,
    ):
        '''
        Init.

        Args:
            encoder:                        is-a `ConvolutionalNeuralNetworks`.
            classifier:                      is-a `NeuralNetworks` or `ConvolutionalNeuralNetworks`.
            pretext_task_model:             is-a `NeuralNetworks` or `ConvolutionalNeuralNetworks`.

            ssda_loss:                      is-a `SSDAloss`.
            initializer:                    is-a `mxnet.initializer.Initializer` for parameters of model. If `None`, it is drawing from the Xavier distribution.
            learning_rate:                  `float` of learning rate.
            learning_attenuate_rate:        `float` of attenuate the `learning_rate` by a factor of this value every `attenuate_epoch`.
            attenuate_epoch:                `int` of attenuate the `learning_rate` by a factor of `learning_attenuate_rate` every `attenuate_epoch`.
                                            
            optimizer_name:                 `str` of name of optimizer.

            ctx:                            `mx.cpu()` or `mx.gpu()`.
            hybridize_flag:                  `bool` of flag that means this class will call `mxnet.gluon.HybridBlock.hybridize()` or not.
            scale:                          `float` of scaling factor for initial parameters.
            tol:                            `float` of the skipping epoch.
                                            If it is not `None`, the epoch will be skipped and parameters will be not updated 
                                            when (loss > previous_loss - tol).

            est:                            `float` of the early stopping.
                                            If it is not `None`, the learning will be stopped 
                                            when (|loss -previous_loss| < est).

        '''
        if isinstance(encoder, ConvolutionalNeuralNetworks) is False:
            raise TypeError("The type of `encoder` must be `ConvolutionalNeuralNetworks`.")

        if isinstance(classifier, NeuralNetworks) is False and isinstance(classifier, ConvolutionalNeuralNetworks) is False:
            raise TypeError("The type of `classifier` must be `NeuralNetworks` or `ConvolutionalNeuralNetworks`.")
        
        if isinstance(pretext_task_model, NeuralNetworks) is False and isinstance(pretext_task_model, ConvolutionalNeuralNetworks) is False:
            raise TypeError("The type of `pretext_task_model` must be `NeuralNetworks` or `ConvolutionalNeuralNetworks`.")

        if isinstance(ssda_loss, SSDALoss) is False:
            raise TypeError("The type of `ssda_loss` must be `SSDALoss`.")

        logger = getLogger("accelbrainbase")
        self.__logger = logger

        super(SSDAController, self).__init__()

        self.encoder = encoder
        self.classifier = classifier
        self.pretext_task_model = pretext_task_model

        self.ssda_loss = ssda_loss

        self.__learning_rate = learning_rate

        self.__ctx = ctx

        self.__loss_list = []
        self.__acc_list = []
        self.__target_domain_arr = None

        self.__tol = tol
        self.__est = est

        self.epoch = 0

    def learn(self, iteratable_data):
        '''
        Learn the observed data points with domain adaptation.

        Args:
            iteratable_data:     is-a `SSDAIterator`.

        '''
        if isinstance(iteratable_data, SSDAIterator) is False:
            raise TypeError("The type of `iteratable_data` must be `SSDAIterator`.")

        self.__loss_list = []
        self.__acc_list = []
        learning_rate = self.__learning_rate
        self.__previous_loss = None
        est_flag = False
        try:
            epoch = self.epoch
            iter_n = 0
            for batch_observed_arr, batch_target_arr, test_batch_observed_arr, test_batch_target_arr, pretext_arr, pretext_label_arr in iteratable_data.generate_learned_samples():
                self.batch_size = batch_observed_arr.shape[0]

                if epoch == 0:
                    source_encoded_arr = self.encoder(batch_observed_arr)
                    target_encoded_arr = self.encoder(pretext_arr)
                    pred_arr = self.classifier(source_encoded_arr)
                    pretext_pred_arr = self.pretext_task_model(target_encoded_arr)

                self.encoder.optimizer.zero_grad()
                self.classifier.optimizer.zero_grad()
                self.pretext_task_model.optimizer.zero_grad()

                source_encoded_arr = self.encoder(batch_observed_arr)
                target_encoded_arr = self.encoder(pretext_arr)
                pred_arr = self.classifier(source_encoded_arr)
                pretext_pred_arr = self.pretext_task_model(target_encoded_arr)

                loss, classification_loss, pretext_loss = self.compute_loss(
                    pretext_pred_arr, 
                    pred_arr, 
                    pretext_label_arr, 
                    batch_target_arr, 
                )
                if self.__previous_loss is not None:
                    loss_diff = loss - self.__previous_loss
                    loss_diff = loss_diff.to('cpu').detach().numpy().copy()
                    if self.__est is not None and np.abs(loss_diff) < self.__est:
                        est_flag = True

                if est_flag is False:
                    loss.backward()
                    self.pretext_task_model.optimizer.step()
                    self.classifier.optimizer.step()
                    self.encoder.optimizer.step()
                    self.regularize()

                    self.__previous_loss = loss

                    if (iter_n+1) % int(iteratable_data.iter_n / iteratable_data.epochs) == 0:
                        with torch.inference_mode():
                            source_encoded_arr = self.encoder.inference(test_batch_observed_arr)
                            target_encoded_arr = self.encoder.inference(pretext_arr)
                            test_pred_arr = self.classifier.inference(source_encoded_arr)
                            pretext_pred_arr = self.pretext_task_model.inference(target_encoded_arr)

                            test_loss, test_classification_loss, test_pretext_loss = self.compute_loss(
                                pretext_pred_arr, 
                                pred_arr, 
                                pretext_label_arr, 
                                batch_target_arr, 
                            )
                        _loss = loss.to('cpu').detach().numpy().copy()
                        _classification_loss = classification_loss.to('cpu').detach().numpy().copy()
                        _pretext_loss = pretext_loss.to('cpu').detach().numpy().copy()
                        _test_loss = test_loss.to('cpu').detach().numpy().copy()
                        _test_classification_loss = test_classification_loss.to('cpu').detach().numpy().copy()
                        _test_pretext_loss = test_pretext_loss.to('cpu').detach().numpy().copy()

                        self.__logger.debug("Epochs: " + str(epoch + 1) + " Train total loss: " + str(_loss) + " Test total loss: " + str(_test_loss))
                        self.__logger.debug("Train classification loss: " + str(_classification_loss) + " Test classification loss: " + str(_test_classification_loss))
                        self.__logger.debug("Train Pretext task loss: " + str(_pretext_loss) + " Test Pretext task loss: " + str(_test_pretext_loss))

                    if self.compute_acc_flag is True:
                        if (iter_n+1) % int(iteratable_data.iter_n / iteratable_data.epochs) == 0:
                            acc, inferenced_label_arr, answer_label_arr = self.compute_acc(pred_arr, batch_target_arr)
                            test_acc, test_inferenced_label_arr, test_answer_label_arr = self.compute_acc(test_pred_arr, test_batch_target_arr)
                            if (epoch + 1) % 100 == 0 or epoch < 100:
                                acc, inferenced_label_arr, answer_label_arr = self.compute_acc(pred_arr, batch_target_arr)
                                test_acc, test_inferenced_label_arr, test_answer_label_arr = self.compute_acc(test_pred_arr, test_batch_target_arr)

                                self.__logger.debug("-" * 100)
                                self.__logger.debug("Train accuracy: " + str(acc) + " Test accuracy: " + str(test_acc))
                                self.__logger.debug("Train infenreced label(inferenced):")
                                self.__logger.debug(inferenced_label_arr)
                                self.__logger.debug("Train infenreced label(answer):")
                                self.__logger.debug(answer_label_arr)

                                self.__logger.debug("Test infenreced label(inferenced):")
                                self.__logger.debug(test_inferenced_label_arr)
                                self.__logger.debug("Test infenreced label(answer):")
                                self.__logger.debug(test_answer_label_arr)
                                self.__logger.debug("-" * 100)

                                if (test_answer_label_arr[0] == test_answer_label_arr).astype(int).sum() != test_answer_label_arr.shape[0]:
                                    if (test_inferenced_label_arr[0] == test_inferenced_label_arr).astype(int).sum() == test_inferenced_label_arr.shape[0]:
                                        self.__logger.debug("It may be overfitting.")

                    if (iter_n+1) % int(iteratable_data.iter_n / iteratable_data.epochs) == 0:
                        self.__loss_list.append(
                            (
                                _loss, 
                                _test_loss,
                                _classification_loss,
                                _test_classification_loss,
                                _pretext_loss,
                            )
                        )
                        if self.compute_acc_flag is True:
                            self.__acc_list.append(
                                (
                                    acc,
                                    test_acc
                                )
                            )

                    if (iter_n+1) % int(iteratable_data.iter_n / iteratable_data.epochs) == 0:
                        epoch += 1
                    iter_n += 1

                else:
                    self.__logger.debug("Early stopping.")
                    self.__logger.debug(str(self.__previous_loss) + " -> " + str(loss_mean))
                    break

        except KeyboardInterrupt:
            self.__logger.debug("Interrupt.")

        self.__logger.debug("end. ")
        self.epoch = epoch

    def inference(self, observed_arr):
        '''
        Inference samples drawn by `IteratableData.generate_inferenced_samples()`.

        Args:
            observed_arr:   rank-2 Array like or sparse matrix as the observed data points.
                            The shape is: (batch size, feature points)

        Returns:
            `mxnet.ndarray` of inferenced feature points.
        '''
        return self(observed_arr)

    def forward(self, x):
        '''
        Hybrid forward with Gluon API.

        Args:
            F:      `mxnet.ndarray` or `mxnet.symbol`.
            x:      `mxnet.ndarray` of observed data points.
        
        Returns:
            `mxnet.ndarray` or `mxnet.symbol` of inferenced feature points.
        '''
        source_encoded_arr = self.encoder(x)
        pred_arr = self.classifier(source_encoded_arr)
        return pred_arr

    def regularize(self):
        '''
        Regularization.
        '''
        self.encoder.regularize()
        self.classifier.regularize()
        self.pretext_task_model.regularize()

    def compute_loss(
        self, 
        pretext_pred_arr, 
        pred_arr, 
        pretext_label_arr, 
        label_arr, 
    ):
        '''
        Compute loss.

        Args:
            pretext_pred_arr:       `mxnet.ndarray` or `mxnet.symbol` of predicted data in pretext, or target domain.
            pred_arr:               `mxnet.ndarray` or `mxnet.symbol` of inferenced labeled feature points in source domain.
            pretext_label_arr:      `mxnet.ndarray` or `mxnet.symbol` of label data in pretext.
            label_arr:              `mxnet.ndarray` or `mxnet.symbol` of label data in source domain.

        Returns:
            Tensor of losses.
        '''
        return self.ssda_loss(
            pretext_pred_arr, 
            pred_arr, 
            pretext_label_arr, 
            label_arr, 
        )

    def compute_acc(self, prob_arr, batch_target_arr):
        '''
        Compute accuracy.

        Args:
            prob_arr:               Softmax probabilities.
            batch_target_arr:       t-hot vectors.
        
        Returns:
            Tuple data.
            - Accuracy.
            - inferenced label.
            - real label.
        '''
        inferenced_label_arr = prob_arr.argmax(axis=1).to('cpu').detach().numpy().copy()
        answer_label_arr = batch_target_arr.argmax(axis=1).to('cpu').detach().numpy().copy()
        acc = (inferenced_label_arr == answer_label_arr).astype(int).sum() / self.batch_size
        return acc, inferenced_label_arr, answer_label_arr

    # `bool`, compute accuracy or not in learning.
    __compute_acc_flag = True

    def get_compute_acc_flag(self):
        ''' getter '''
        return self.__compute_acc_flag
    
    def set_compute_acc_flag(self, value):
        ''' setter '''
        self.__compute_acc_flag = value

    compute_acc_flag = property(get_compute_acc_flag, set_compute_acc_flag)

    def set_readonly(self, value):
        ''' setter '''
        raise TypeError("This property must be read-only.")

    def get_loss_arr(self):
        ''' getter for losses. '''
        return np.array(self.__loss_list)

    loss_arr = property(get_loss_arr, set_readonly)

    def get_acc_list(self):
        ''' getter for accuracies. '''
        return np.array(self.__acc_list)
    
    acc_arr = property(get_acc_list, set_readonly)



# -*- coding: utf-8 -*-
from accelbrainbase.controllable_model import ControllableModel
from accelbrainbase.computable_loss import ComputableLoss

from accelbrainbase.iteratabledata.ssda_transformer_iterator import SSDATransformerIterator

from accelbrainbase.observabledata._torch.neural_networks import NeuralNetworks
from accelbrainbase.observabledata._torch.convolutional_neural_networks import ConvolutionalNeuralNetworks
from accelbrainbase.controllablemodel._torch.transformer_controller import TransformerController
import numpy as np
from logging import getLogger
import torch
from torch import nn


class SSDATransformerController(nn.Module, ControllableModel):
    '''
    Self-supervised domain adaptive transformers.
    
    This library dryly considers the various Transformers variants 
    such as BERT, XLNet, RoBERTa, ALBERT, etc, 
    are merely applications of "self-supervised learning"
    or "self-supervised domain adaptation(SSDA)."

    From this point of view, this class builds 
    the Transformers variants as SSDA models.

    Of course, there is no necessity to build 
    an existing model like BERT or ALBERT as it is.

    This class allows you to build any combination of 
    transformers models that solve downstream tasks 
    and different models that solve pretext tasks.

    References:
        - Jing, L., & Tian, Y. (2020). Self-supervised visual feature learning with deep neural networks: A survey. IEEE Transactions on Pattern Analysis and Machine Intelligence.
        - Xu, J., Xiao, L., & López, A. M. (2019). Self-supervised domain adaptation for computer vision tasks. IEEE Access, 7, 156694-156706., p156698.

    '''
    
    def __init__(
        self,
        transformer_controller,
        downstream_task_model,
        downstream_task_loss,
        pretext_task_model_list,
        pretext_task_loss_list,
        learning_rate=1e-05,
        ctx="cpu",
        scale=1.0,
        tol=3.0,
        est=1e-08,
        wd=0.0,
    ):
        '''
        Init.

        Args:
            transformer_controller:         is-a `TransformerController`.
            downstream_task_model:          is-a `NeuralNetworks` or `ConvolutionalNeuralNetworks`.
            downstream_task_loss:           is-a 
            pretext_task_model_list:        `list` of `NeuralNetworks`s or `ConvolutionalNeuralNetworks`s.

            initializer:                    is-a `mxnet.initializer.Initializer` for parameters of model. If `None`, it is drawing from the Xavier distribution.
            learning_rate:                  `float` of learning rate.
            learning_attenuate_rate:        `float` of attenuate the `learning_rate` by a factor of this value every `attenuate_epoch`.
            attenuate_epoch:                `int` of attenuate the `learning_rate` by a factor of `learning_attenuate_rate` every `attenuate_epoch`.
                                            
            optimizer_name:                 `str` of name of optimizer.

            ctx:                            `mx.cpu()` or `mx.gpu()`.
            hybridize_flag:                  `bool` of flag that means this class will call `mxnet.gluon.HybridBlock.hybridize()` or not.
            scale:                          `float` of scaling factor for initial parameters.
            tol:                            `float` of the skipping epoch.
                                            If it is not `None`, the epoch will be skipped and parameters will be not updated 
                                            when (loss > previous_loss - tol).

            est:                            `float` of the early stopping.
                                            If it is not `None`, the learning will be stopped 
                                            when (|loss -previous_loss| < est).

        '''
        if isinstance(transformer_controller, TransformerController) is False:
            raise TypeError("The type of `transformer_controller` must be `TransformerController`.")

        if isinstance(downstream_task_model, NeuralNetworks) is False and isinstance(downstream_task_model, ConvolutionalNeuralNetworks) is False:
            raise TypeError("The type of `downstream_task_model` must be `NeuralNetworks` or `ConvolutionalNeuralNetworks`.")

        if isinstance(downstream_task_loss, ComputableLoss) is False and isinstance(downstream_task_loss, nn.modules.loss._Loss) is False:
            raise TypeError("The type of `downstream_task_loss` must be `ComputableLoss` or `nn.modules.loss._Loss`.")

        for pretext_task_model in pretext_task_model_list:
            if isinstance(pretext_task_model, NeuralNetworks) is False and isinstance(pretext_task_model, ConvolutionalNeuralNetworks) is False:
                raise TypeError("The type of value in `pretext_task_model_list` must be `NeuralNetworks` or `ConvolutionalNeuralNetworks`.")

        for pretext_task_loss in pretext_task_loss_list:
            if isinstance(pretext_task_loss, ComputableLoss) is False and isinstance(pretext_task_loss, nn.modules.loss._Loss) is False:
                raise TypeError("The type of `pretext_task_loss` must be `ComputableLoss` or `nn.modules.loss._Loss`.")

        logger = getLogger("accelbrainbase")
        self.__logger = logger

        super(SSDATransformerController, self).__init__()

        self.transformer_controller = transformer_controller
        self.downstream_task_model = downstream_task_model
        self.downstream_task_loss = downstream_task_loss
        self.pretext_task_model_list = nn.ModuleList(pretext_task_model_list)
        self.pretext_task_loss_list = pretext_task_loss_list

        self.__learning_rate = learning_rate

        self.__ctx = ctx

        self.__loss_list = []
        self.__acc_list = []
        self.__target_domain_arr = None

        self.__tol = tol
        self.__est = est

        self.epoch = 0

    def learn(self, iteratable_data):
        '''
        Learn the observed data points with domain adaptation.

        Args:
            iteratable_data:     is-a `SSDATransformerIterator`.

        '''
        if isinstance(iteratable_data, SSDATransformerIterator) is False:
            raise TypeError("The type of `iteratable_data` must be `SSDATransformerIterator`.")

        self.__loss_list = []
        self.__acc_list = []
        learning_rate = self.__learning_rate
        self.__previous_loss = None
        est_flag = False
        try:
            epoch = self.epoch
            iter_n = 0
            for batch_data_tuple in iteratable_data.generate_learned_samples():
                (
                    encoded_observed_arr, 
                    decoded_observed_arr, 
                    encoded_mask_arr, 
                    decoded_mask_arr, 
                    test_encoded_observed_arr, 
                    test_decoded_observed_arr, 
                    test_encoded_mask_arr, 
                    test_decoded_mask_arr, 
                    training_target_arr, 
                    test_target_arr, 
                    pretext_encoded_observed_arr_list, 
                    pretext_decoded_observed_arr_list, 
                    pretext_encoded_mask_arr_list, 
                    pretext_decoded_mask_arr_list, 
                    pretext_label_arr_list
                ) = batch_data_tuple
                self.batch_size = encoded_observed_arr.shape[0]

                if epoch == 0:
                    source_encoded_arr = self.transformer_controller(
                        encoded_observed_arr, 
                        decoded_observed_arr, 
                        encoded_mask_arr, 
                        decoded_mask_arr, 
                    )
                    _ = self.downstream_task_model(source_encoded_arr)
                    target_encoded_arr_list = []
                    for i in range(len(pretext_label_arr_list)):
                        target_encoded_arr = self.transformer_controller(
                            pretext_encoded_observed_arr_list[i], 
                            pretext_decoded_observed_arr_list[i], 
                            pretext_encoded_mask_arr_list[i], 
                            pretext_decoded_mask_arr_list[i], 
                        )
                        target_encoded_arr_list.append(target_encoded_arr)

                    pred_arr = self.downstream_task_model(source_encoded_arr)
                    pretext_pred_arr_list = []
                    for i in range(len(target_encoded_arr_list)):
                        pretext_pred_arr = self.pretext_task_model_list[i](
                            target_encoded_arr_list[i]
                        )
                        pretext_pred_arr_list.append(pretext_pred_arr)

                self.transformer_controller.optimizer.zero_grad()
                self.transformer_controller.encoder.optimizer.zero_grad()
                self.transformer_controller.decoder.optimizer.zero_grad()
                self.downstream_task_model.optimizer.zero_grad()
                for i in range(len(self.pretext_task_model_list)):
                    self.pretext_task_model_list[i].optimizer.zero_grad()

                source_encoded_arr = self.transformer_controller(
                    encoded_observed_arr, 
                    decoded_observed_arr, 
                    encoded_mask_arr, 
                    decoded_mask_arr, 
                )
                target_encoded_arr_list = []
                for i in range(len(pretext_encoded_observed_arr_list)):
                    target_encoded_arr = self.transformer_controller(
                        pretext_encoded_observed_arr_list[i], 
                        pretext_decoded_observed_arr_list[i], 
                        pretext_encoded_mask_arr_list[i], 
                        pretext_decoded_mask_arr_list[i], 
                    )
                    target_encoded_arr_list.append(target_encoded_arr)

                pred_arr = self.downstream_task_model(source_encoded_arr)

                pretext_pred_arr_list = []
                for i in range(len(target_encoded_arr_list)):
                    pretext_pred_arr = self.pretext_task_model_list[i](
                        target_encoded_arr_list[i]
                    )
                    pretext_pred_arr_list.append(pretext_pred_arr)

                pretext_loss = 0.0
                for i in range(len(self.pretext_task_loss_list)):
                    pretext_loss = pretext_loss + self.pretext_task_loss_list[i](
                        pretext_pred_arr_list[i],
                        pretext_label_arr_list[i]
                    )

                downstream_task_loss = self.downstream_task_loss(
                    pred_arr,
                    training_target_arr
                )

                loss = downstream_task_loss + pretext_loss

                if self.__previous_loss is not None:
                    loss_diff = loss - self.__previous_loss
                    loss_diff = loss_diff.to('cpu').detach().numpy().copy()
                    if self.__est is not None and np.abs(loss_diff) < self.__est:
                        est_flag = True

                if est_flag is False:
                    loss.backward()
                    for i in range(len(self.pretext_task_model_list)):
                        self.pretext_task_model_list[i].optimizer.step()
                    self.downstream_task_model.optimizer.step()
                    self.transformer_controller.optimizer.step()
                    self.transformer_controller.decoder.optimizer.step()
                    self.transformer_controller.encoder.optimizer.step()
                    self.regularize()

                    self.__previous_loss = loss

                    if (iter_n+1) % int(iteratable_data.iter_n / iteratable_data.epochs) == 0:
                        with torch.inference_mode():
                            source_encoded_arr = self.transformer_controller(
                                test_encoded_observed_arr, 
                                test_decoded_observed_arr, 
                                test_encoded_mask_arr, 
                                test_decoded_mask_arr, 
                            )
                            target_encoded_arr_list = []
                            for i in range(len(pretext_encoded_observed_arr_list)):
                                target_encoded_arr = self.transformer_controller(
                                    pretext_encoded_observed_arr_list[i], 
                                    pretext_decoded_observed_arr_list[i], 
                                    pretext_encoded_mask_arr_list[i], 
                                    pretext_decoded_mask_arr_list[i], 
                                )
                                target_encoded_arr_list.append(target_encoded_arr)

                            test_pred_arr = self.downstream_task_model(source_encoded_arr)
                            pretext_pred_arr_list = []
                            for i in range(len(target_encoded_arr_list)):
                                pretext_pred_arr = self.pretext_task_model_list[i](
                                    target_encoded_arr_list[i]
                                )
                                pretext_pred_arr_list.append(pretext_pred_arr)

                            test_pretext_loss = 0.0
                            for i in range(len(self.pretext_task_loss_list)):
                                test_pretext_loss = pretext_loss + self.pretext_task_loss_list[i](
                                    pretext_pred_arr_list[i],
                                    pretext_label_arr_list[i]
                                )

                            test_downstream_task_loss = self.downstream_task_loss(
                                test_pred_arr,
                                test_target_arr
                            )

                            test_loss = test_downstream_task_loss + test_pretext_loss

                        _loss = loss.to('cpu').detach().numpy().copy()
                        _downstream_task_loss = downstream_task_loss.to('cpu').detach().numpy().copy()
                        _pretext_loss = pretext_loss.to('cpu').detach().numpy().copy()
                        _test_loss = test_loss.to('cpu').detach().numpy().copy()
                        _test_downstream_task_loss = test_downstream_task_loss.to('cpu').detach().numpy().copy()
                        _test_pretext_loss = test_pretext_loss.to('cpu').detach().numpy().copy()

                        self.__logger.debug("Epochs: " + str(epoch + 1) + " Train total loss: " + str(_loss) + " Test total loss: " + str(_test_loss))
                        self.__logger.debug("Train classification loss: " + str(_downstream_task_loss) + " Test classification loss: " + str(_test_downstream_task_loss))
                        self.__logger.debug("Train Pretext task loss: " + str(_pretext_loss) + " Test Pretext task loss: " + str(_test_pretext_loss))

                    if self.compute_acc_flag is True:
                        if (iter_n+1) % int(iteratable_data.iter_n / iteratable_data.epochs) == 0:
                            acc, inferenced_label_arr, answer_label_arr = self.compute_acc(pred_arr, training_target_arr)
                            test_acc, test_inferenced_label_arr, test_answer_label_arr = self.compute_acc(test_pred_arr, test_target_arr)
                            if (epoch + 1) % 100 == 0 or epoch < 100:
                                acc, inferenced_label_arr, answer_label_arr = self.compute_acc(pred_arr, training_target_arr)
                                test_acc, test_inferenced_label_arr, test_answer_label_arr = self.compute_acc(test_pred_arr, test_target_arr)

                                self.__logger.debug("-" * 100)
                                self.__logger.debug("Train accuracy: " + str(acc) + " Test accuracy: " + str(test_acc))
                                self.__logger.debug("Train infenreced label(inferenced):")
                                self.__logger.debug(inferenced_label_arr)
                                self.__logger.debug("Train infenreced label(answer):")
                                self.__logger.debug(answer_label_arr)

                                self.__logger.debug("Test infenreced label(inferenced):")
                                self.__logger.debug(test_inferenced_label_arr)
                                self.__logger.debug("Test infenreced label(answer):")
                                self.__logger.debug(test_answer_label_arr)
                                self.__logger.debug("-" * 100)

                                if (test_answer_label_arr[0] == test_answer_label_arr).astype(int).sum() != test_answer_label_arr.shape[0]:
                                    if (test_inferenced_label_arr[0] == test_inferenced_label_arr).astype(int).sum() == test_inferenced_label_arr.shape[0]:
                                        self.__logger.debug("It may be overfitting.")

                    if (iter_n+1) % int(iteratable_data.iter_n / iteratable_data.epochs) == 0:
                        self.__loss_list.append(
                            (
                                _loss, 
                                _test_loss,
                                _downstream_task_loss,
                                _test_downstream_task_loss,
                                _pretext_loss,
                            )
                        )
                        if self.compute_acc_flag is True:
                            self.__acc_list.append(
                                (
                                    acc,
                                    test_acc
                                )
                            )

                    if (iter_n+1) % int(iteratable_data.iter_n / iteratable_data.epochs) == 0:
                        epoch += 1
                    iter_n += 1

                else:
                    self.__logger.debug("Early stopping.")
                    self.__logger.debug(str(self.__previous_loss) + " -> " + str(loss_mean))
                    break

        except KeyboardInterrupt:
            self.__logger.debug("Interrupt.")

        self.__logger.debug("end. ")
        self.epoch = epoch

    def inference(
        self, 
        encoded_observed_arr, 
        decoded_observed_arr, 
        encoded_mask_arr=None, 
        decoded_mask_arr=None, 
    ):
        '''
        Inference samples drawn by `IteratableData.generate_inferenced_samples()`.

        Args:
            observed_arr:   rank-2 Array like or sparse matrix as the observed data points.
                            The shape is: (batch size, feature points)

        Returns:
            `mxnet.ndarray` of inferenced feature points.
        '''
        return self(
            encoded_observed_arr, 
            decoded_observed_arr, 
            encoded_mask_arr, 
            decoded_mask_arr, 
        )

    def forward(
        self, 
        encoded_observed_arr, 
        decoded_observed_arr, 
        encoded_mask_arr, 
        decoded_mask_arr, 
    ):
        '''
        Hybrid forward with Gluon API.

        Args:
            F:      `mxnet.ndarray` or `mxnet.symbol`.
            x:      `mxnet.ndarray` of observed data points.
        
        Returns:
            `mxnet.ndarray` or `mxnet.symbol` of inferenced feature points.
        '''
        source_encoded_arr = self.transformer_controller(
            encoded_observed_arr, 
            decoded_observed_arr, 
            encoded_mask_arr, 
            decoded_mask_arr, 
        )
        pred_arr = self.downstream_task_model(source_encoded_arr)
        return pred_arr

    def regularize(self):
        '''
        Regularization.
        '''
        self.transformer_controller.regularize()
        self.downstream_task_model.regularize()
        for i in range(len(self.pretext_task_model_list)):
            self.pretext_task_model_list[i].regularize()

    def compute_loss(
        self, 
        pred_arr, 
        label_arr, 
    ):
        '''
        Compute loss.

        Args:
            pretext_pred_arr:       `mxnet.ndarray` or `mxnet.symbol` of predicted data in pretext, or target domain.
            pred_arr:               `mxnet.ndarray` or `mxnet.symbol` of inferenced labeled feature points in source domain.
            pretext_label_arr:      `mxnet.ndarray` or `mxnet.symbol` of label data in pretext.
            label_arr:              `mxnet.ndarray` or `mxnet.symbol` of label data in source domain.

        Returns:
            Tensor of losses.
        '''
        return self.downstream_task_loss(
            pred_arr, 
            label_arr, 
        )

    def compute_acc(self, prob_arr, batch_target_arr):
        '''
        Compute accuracy.

        Args:
            prob_arr:               Softmax probabilities.
            batch_target_arr:       t-hot vectors.
        
        Returns:
            Tuple data.
            - Accuracy.
            - inferenced label.
            - real label.
        '''
        inferenced_label_arr = prob_arr.argmax(axis=1).to('cpu').detach().numpy().copy()
        answer_label_arr = batch_target_arr.argmax(axis=1).to('cpu').detach().numpy().copy()
        acc = (inferenced_label_arr == answer_label_arr).astype(int).sum() / self.batch_size
        return acc, inferenced_label_arr, answer_label_arr

    # `bool`, compute accuracy or not in learning.
    __compute_acc_flag = True

    def get_compute_acc_flag(self):
        ''' getter '''
        return self.__compute_acc_flag
    
    def set_compute_acc_flag(self, value):
        ''' setter '''
        self.__compute_acc_flag = value

    compute_acc_flag = property(get_compute_acc_flag, set_compute_acc_flag)

    def set_readonly(self, value):
        ''' setter '''
        raise TypeError("This property must be read-only.")

    def get_loss_arr(self):
        ''' getter for losses. '''
        return np.array(self.__loss_list)

    loss_arr = property(get_loss_arr, set_readonly)

    def get_acc_list(self):
        ''' getter for accuracies. '''
        return np.array(self.__acc_list)
    
    acc_arr = property(get_acc_list, set_readonly)



# -*- coding: utf-8 -*-
import numpy as np
import torch

import pandas as pd
from logging import getLogger
import os

from accelbrainbase.extractabledata.image_extractor import ImageExtractor
from accelbrainbase.iteratabledata.labeled_image_iterator import LabeledImageIterator as _LabeledImageIterator
from accelbrainbase.noiseable_data import NoiseableData


class LabeledVideoIterator(_LabeledImageIterator):
    '''
    Iterator that draws from image files and generates `mxnet.ndarray` of labeled samples.
    '''

    def __init__(
        self,
        image_extractor,
        dir_list,
        test_dir_list=None,
        epochs=300,
        batch_size=20,
        seq_len=10,
        at_intervals=1,
        norm_mode="z_score",
        scale=1.0,
        noiseable_data=None,
        ctx="cpu"
    ):
        '''
        Init.

        Args:
            image_extractor:                is-a `ImageExtractor`.
            dir_list:                       `list` of directories that store your image files in training.
                                            This class will not scan the directories recursively and consider that
                                            all image file will be sorted by any rule in relation to your sequencial modeling.

            test_dir_list:                  `list` of directories that store your image files in test.
                                            If `None`, this value will be equivalent to `dir_list`.
                                            This class will not scan the directories recursively and consider that
                                            all image file will be sorted by any rule in relation to your sequencial modeling.

            epochs:                         `int` of epochs of Mini-batch.
            bath_size:                      `int` of batch size of Mini-batch.
            seq_len:                        `int` of the length of series.
            at_intervals:                   `int` of intervals in one sequence.

            norm_mode:                      How to normalize pixel values of images.
                                            - `z_score`: Z-Score normalization.
                                            - `min_max`: Min-max normalization.
                                            - others : This class will not normalize the data.

            scale:                          `float` of scaling factor for data.
            noiseable_data:                 is-a `NoiseableData` for Denoising Auto-Encoders.
            ctx:                            `mx.cpu()` or `mx.gpu()`.
        '''

        if isinstance(image_extractor, ImageExtractor) is False:
            raise TypeError("The type of `image_extractor` must be `ImageExtractor`.")
        if isinstance(dir_list, list) is False:
            raise TypeError("The type of `dir_list` must be `list`.")
        if noiseable_data is not None and isinstance(noiseable_data, NoiseableData) is False:
            raise TypeError("The type of `noiseable_data` must be `NoiseableData`.")
        if isinstance(seq_len, int) is False:
            raise TypeError("The type of `seq_len` must be `int`.")
        if seq_len <= 0:
            raise ValueError("The value of `seq_len` must be more than `0`.")
        if isinstance(at_intervals, int) is False:
            raise TypeError("The type of `at_intervals` must be `int`.")
        if at_intervals <= 0:
            raise ValueError("The value of `at_intervals` must be more than `0`.")

        logger = getLogger("accelbrainbase")
        self.__logger = logger

        dir_list.sort()
        self.__training_file_path_list = [None] * len(dir_list)
        dataset_size = 0
        for i in range(len(dir_list)):
            file_path_list = [dir_list[i] + "/" + file_name for file_name in os.listdir(dir_list[i] + "/")]
            file_path_list.sort()
            self.__training_file_path_list[i] = file_path_list
            dataset_size = dataset_size + len(file_path_list)

        iter_n = int(epochs * max(dataset_size / batch_size, 1))

        if test_dir_list is not None and isinstance(test_dir_list, list) is False:
            raise TypeError("The type of `test_dir_list` must be `list`.")
        elif test_dir_list is None:
            test_dir_list = dir_list
            self.__test_file_path_list = self.__training_file_path_list
        else:
            test_dir_list.sort()
            self.__test_file_path_list = [None] * len(test_dir_list)
            for i in range(len(test_dir_list)):
                file_path_list = [test_dir_list[i] + "/" + file_name for file_name in os.listdir(test_dir_list[i] + "/")]
                file_path_list.sort()
                self.__test_file_path_list[i] = file_path_list

        self.__image_extractor = image_extractor
        self.__dir_list = dir_list
        self.__test_dir_list = test_dir_list

        self.iter_n = iter_n
        self.epochs = epochs
        self.batch_size = batch_size
        self.__seq_len = seq_len
        self.__at_intervals = at_intervals
        self.norm_mode = norm_mode
        self.scale = scale
        self.__noiseable_data = noiseable_data
        self.__ctx = ctx

    def generate_learned_samples(self):
        '''
        Draw and generate data.

        Returns:
            `Tuple` data. The shape is ...
            - `mxnet.ndarray` of observed data points in training.
            - `mxnet.ndarray` of supervised data in training.
            - `mxnet.ndarray` of observed data points in test.
            - `mxnet.ndarray` of supervised data in test.
        '''
        for _ in range(self.iter_n):
            training_batch_arr, test_batch_arr = None, None
            training_label_arr, test_label_arr = None, None
            for batch_size in range(self.batch_size):
                dir_key = np.random.randint(low=0, high=len(self.__training_file_path_list))

                training_one_hot_arr = torch.zeros((1, len(self.__training_file_path_list)), device=self.__ctx)
                training_one_hot_arr[0, dir_key] = 1

                training_file_path_list = self.__split_at_intervals(
                    self.__training_file_path_list[dir_key], 
                    start_pos=0, 
                    seq_interval=self.__at_intervals
                )

                training_data_arr, test_data_arr = None, None
                training_file_key = np.random.randint(
                    low=0,
                    high=len(training_file_path_list) - self.__seq_len
                )

                test_dir_key = np.random.randint(low=0, high=len(self.__test_file_path_list))
                test_one_hot_arr = torch.zeros((1, len(self.__test_file_path_list)), device=self.__ctx)
                test_one_hot_arr[0, test_dir_key] = 1

                test_file_path_list = self.__split_at_intervals(
                    self.__test_file_path_list[test_dir_key], 
                    start_pos=0, 
                    seq_interval=self.__at_intervals
                )

                test_file_key = np.random.randint(
                    low=0,
                    high=len(test_file_path_list) - self.__seq_len
                )
                for seq in range(self.__seq_len):
                    seq_training_batch_arr = self.__image_extractor.extract(
                        path=training_file_path_list[training_file_key+seq],
                    )
                    seq_training_batch_arr = self.pre_normalize(seq_training_batch_arr)
                    seq_training_batch_arr = torch.unsqueeze(seq_training_batch_arr, dim=0)
                    seq_test_batch_arr = self.__image_extractor.extract(
                        path=test_file_path_list[test_file_key+seq],
                    )
                    seq_test_batch_arr = self.pre_normalize(seq_test_batch_arr)
                    seq_test_batch_arr = torch.unsqueeze(seq_test_batch_arr, dim=0)

                    if training_data_arr is not None:
                        training_data_arr = torch.cat((training_data_arr, seq_training_batch_arr), dim=0)
                    else:
                        training_data_arr = seq_training_batch_arr
                    
                    if test_data_arr is not None:
                        test_data_arr = torch.cat((test_data_arr, seq_test_batch_arr), dim=0)
                    else:
                        test_data_arr = seq_test_batch_arr

                training_data_arr = torch.unsqueeze(training_data_arr, dim=0)
                test_data_arr = torch.unsqueeze(test_data_arr, dim=0)

                if training_batch_arr is not None:
                    training_batch_arr = torch.cat((training_batch_arr, training_data_arr), dim=0)
                else:
                    training_batch_arr = training_data_arr

                if test_batch_arr is not None:
                    test_batch_arr = torch.cat((test_batch_arr, test_data_arr), dim=0)
                else:
                    test_batch_arr = test_data_arr

                if training_label_arr is not None:
                    training_label_arr = torch.cat((training_label_arr, training_one_hot_arr), dim=0)
                else:
                    training_label_arr = training_one_hot_arr

                if test_label_arr is not None:
                    test_label_arr = torch.cat((test_label_arr, test_one_hot_arr), dim=0)
                else:
                    test_label_arr = test_one_hot_arr

            if self.__noiseable_data is not None:
                training_batch_arr = self.__noiseable_data.noise(training_batch_arr)

            yield training_batch_arr, training_label_arr, test_batch_arr, test_label_arr

    def pre_normalize(self, arr):
        '''
        Normalize before observation.

        Args:
            arr:    Tensor.
        
        Returns:
            Tensor.
        '''
        if self.norm_mode == "min_max":
            if torch.max(arr) != torch.min(arr):
                n = 0.0
            else:
                n = 1e-08
            arr = (arr - torch.min(arr)) / (torch.max(arr) - torch.min(arr) + n)
        elif self.norm_mode == "z_score":
            std = torch.std(arr)
            if std == 0:
                std += 1e-08
            arr = (arr - torch.mean(arr)) / std

        arr = arr * self.scale
        return arr

    def generate_inferenced_samples(self):
        '''
        Draw and generate data.
        The targets will be drawn from all image file sorted in ascending order by file name.

        Returns:
            `Tuple` data. The shape is ...
            - `None`.
            - `int` of key of directory.
            - `mxnet.ndarray` of observed data points in test.
            - file path.
        '''
        scan_file_path_list = []
        for dir_key in range(len(self.__test_file_path_list)):
            for file_key in range(len(self.__test_file_path_list[dir_key]) - self.__seq_len):
                _scan_file_path_list = []
                for seq in range(self.__seq_len):
                    file_path = self.__test_file_path_list[dir_key][file_key+seq]
                    _scan_file_path_list.append(file_path)
                scan_file_path_list.append(_scan_file_path_list)

        import random
        random.shuffle(scan_file_path_list)

        test_batch_arr = None
        file_path_list = []
        for _file_path_list_ in scan_file_path_list:
            test_data_arr = None
            for seq in range(self.__seq_len):
                file_path = _file_path_list_[seq]
                seq_test_batch_arr = self.__image_extractor.extract(
                    path=file_path,
                )
                seq_test_batch_arr = self.pre_normalize(seq_test_batch_arr)
                seq_test_batch_arr = torch.unsqueeze(seq_test_batch_arr, dim=0)
                if test_data_arr is not None:
                    test_data_arr = torch.cat((test_data_arr, seq_test_batch_arr), dim=0)
                else:
                    test_data_arr = seq_test_batch_arr

            test_data_arr = torch.unsqueeze(test_data_arr, dim=0)
            if test_batch_arr is not None:
                test_batch_arr = torch.cat((test_batch_arr, test_data_arr), dim=0)
            else:
                test_batch_arr = test_data_arr

            file_path_list.append(file_path)

            if test_batch_arr.shape[0] == self.batch_size:
                test_batch_arr_ = test_batch_arr
                test_batch_arr = None
                _file_path_list = file_path_list
                file_path_list = []
                yield None, None, test_batch_arr_, _file_path_list
            elif test_batch_arr.shape[0] > self.batch_size:
                raise ValueError()

    def __split_at_intervals(self, file_path_list, start_pos=0, seq_interval=1):
        file_path_arr = np.array(file_path_list)
        return file_path_arr[list(range(start_pos, len(file_path_list), seq_interval))].tolist()



# -*- coding: utf-8 -*-
from accelbrainbase.samplabledata.pretext_sampler import PretextSampler
import numpy as np
import torch


class MaskedLanguageSampler(PretextSampler):
    '''
    Sampler for pretext task based on the Masked Language Modeling.

    References:
        - Devlin, J., Chang, M. W., Lee, K., & Toutanova, K. (2018). Bert: Pre-training of deep bidirectional transformers for language understanding. arXiv preprint arXiv:1810.04805.
        - Taylor, W. L. (1953). “Cloze procedure”: A new tool for measuring readability. Journalism quarterly, 30(4), 415-433.
    '''

    def __init__(
        self,
        masked_symbol=-1000,
    ):
        '''
        Init.

        Args:
            masked_symbol:      Symbol of mask.
        '''
        self.masked_symbol = masked_symbol

    def preprocess(self, target_domain_arr):
        '''
        Preprocess observed data points in target domain.

        Args:
            target_domain_arr:      Tensor of observed data points in target domain.
                                    The rank of tensor is more than 2.
                                    The shape is (batch, sequence, token vector, ...).
        '''
        pretext_label_arr = None
        target_domain_arr = target_domain_arr.float()
        pretext_encoded_observed_arr = target_domain_arr.detach()
        pretext_decoded_observed_arr = torch.ones_like(target_domain_arr) * self.masked_symbol
        for batch in range(target_domain_arr.shape[0]):
            target_seq = np.random.randint(
                low=0,
                high=target_domain_arr[batch].shape[0]
            )
            if pretext_label_arr is None:
                pretext_label_arr = torch.unsqueeze(
                    target_domain_arr[batch, target_seq],
                    axis=0
                )
            else:
                pretext_label_arr = torch.cat(
                    (
                        pretext_label_arr,
                        torch.unsqueeze(
                            target_domain_arr[batch, target_seq],
                            axis=0
                        )
                    ),
                    dim=0
                )
            target_domain_arr[batch, target_seq] = self.masked_symbol

        self.pretext_encoded_observed_arr = target_domain_arr.float()
        self.pretext_decoded_observed_arr = pretext_decoded_observed_arr.float()
        self.pretext_encoded_mask_arr = None
        self.pretext_decoded_mask_arr = None
        self.pretext_label_arr = pretext_label_arr.float()



# -*- coding: utf-8 -*-
from accelbrainbase.controllable_model import ControllableModel
from accelbrainbase._mxnet._exception.init_deferred_error import InitDeferredError
from accelbrainbase.observabledata._mxnet.transformer_model import TransformerModel
from accelbrainbase.observabledata._mxnet.transformermodel.transformer_encoder import TransformerEncoder
from accelbrainbase.observabledata._mxnet.transformermodel.transformer_decoder import TransformerDecoder
from accelbrainbase.iteratabledata.transformer_iterator import TransformerIterator
from accelbrainbase.computable_loss import ComputableLoss
from accelbrainbase.regularizatable_data import RegularizatableData

from mxnet.gluon.block import HybridBlock
from mxnet import gluon
from mxnet import autograd
import numpy as np
import mxnet as mx
import mxnet.ndarray as nd
from mxnet import MXNetError
from logging import getLogger


class TransformerController(HybridBlock, ControllableModel):
    '''
    Transformer.

    References:
        - Bahdanau, D., Cho, K., & Bengio, Y. (2014). Neural machine translation by jointly learning to align and translate. arXiv preprint arXiv:1409.0473.
        - Floridi, L., & Chiriatti, M. (2020). GPT-3: Its nature, scope, limits, and consequences. Minds and Machines, 30(4), 681-694.
        - Miller, A., Fisch, A., Dodge, J., Karimi, A. H., Bordes, A., & Weston, J. (2016). Key-value memory networks for directly reading documents. arXiv preprint arXiv:1606.03126.
        - Radford, A., Narasimhan, K., Salimans, T., & Sutskever, I. (2018) Improving Language Understanding by Generative Pre-Training. OpenAI (URL: https://s3-us-west-2.amazonaws.com/openai-assets/research-covers/language-unsupervised/language_understanding_paper.pdf)
        - Radford, A., Wu, J., Child, R., Luan, D., Amodei, D., & Sutskever, I. (2019). Language models are unsupervised multitask learners. OpenAI blog, 1(8), 9.
        - Vaswani, A., Shazeer, N., Parmar, N., Uszkoreit, J., Jones, L., Gomez, A. N., & Polosukhin, I. (2017). Attention is all you need. arXiv preprint arXiv:1706.03762.

    '''
    # `bool` that means initialization in this class will be deferred or not.
    __init_deferred_flag = False

    def __init__(
        self,
        computable_loss=None,
        encoder=None,
        decoder=None,
        layer_n=3,
        head_n=3,
        seq_len=5,
        depth_dim=100,
        hidden_dim=100,
        self_attention_activation_list=[],
        multi_head_attention_activation_list=[],
        fc_activation_list=[],
        optimizer_name="SGD",
        learning_rate=1e-05,
        learning_attenuate_rate=1.0,
        attenuate_epoch=50,
        dropout_rate=0.5,
        hybridize_flag=True,
        ctx=mx.gpu(),
        initializer=None,
        regularizatable_data_list=[],
        weight_decay=0.01,
        positional_embedding_weignt=1.0,
        **kwargs
    ):
        '''
        Init.

        Args:
            computable_loss:                is-a `ComputableLoss` or `gluon.loss`.
            encoder:                        is-a `TransformerModel`.
            decoder:                        is-a `TransformerModel`.
            layer_n:                        `int` of the number of layers.
            head_n:                         `int` of the number of heads for multi-head attention model.
            seq_len:                        `int` of the length of sequences.
            depth_dim:                      `int` of dimension of dense layer.
            hidden_dim:                     `int` of dimension of hidden(encoder) layer.
            self_attention_activation_list: `list` of `str` of activation function for self-attention model.
            multi_head_attention_activation_list:   `list` of `str` of activation function for multi-head attention model.
            fc_activation_list:             `list` of `str` of activation function in fully-connected layers.
            learning_rate:                  `float` of learning rate.
            learning_attenuate_rate:        `float` of attenuate the `learning_rate` by a factor of this value every `attenuate_epoch`.
            attenuate_epoch:                `int` of attenuate the `learning_rate` by a factor of `learning_attenuate_rate` every `attenuate_epoch`.
            optimizer_name:                 `str` of name of optimizer.
            hybridize_flag:                  Call `mxnet.gluon.HybridBlock.hybridize()` or not.
            scale:                          `float` of scaling factor for initial parameters.
            ctx:                            `mx.cpu()` or `mx.gpu()`.
            initializer:                    is-a `mxnet.initializer` for parameters of model. If `None`, it is drawing from the Xavier distribution.

        '''
        super(TransformerController, self).__init__()

        if computable_loss is None:
            computable_loss = gluon.loss.SoftmaxCrossEntropyLoss(
                axis=-1, 
                sparse_label=False, 
                from_logits=False, 
                weight=None, 
                batch_axis=0
            )
        self.__computable_loss = computable_loss

        if encoder is None:
            if hidden_dim is None or hidden_dim == depth_dim:
                encoder = TransformerEncoder(
                    depth_dim=depth_dim,
                    layer_n=layer_n,
                    head_n=head_n,
                    self_attention_activation_list=self_attention_activation_list,
                    fc_activation_list=fc_activation_list,
                    computable_loss=computable_loss,
                    initializer=initializer,
                    not_init_flag=True,
                    hybridize_flag=hybridize_flag,
                    dropout_rate=dropout_rate
                )
            else:
                encoder = TransformerEncoder(
                    depth_dim=hidden_dim,
                    layer_n=layer_n,
                    head_n=head_n,
                    self_attention_activation_list=self_attention_activation_list,
                    fc_activation_list=fc_activation_list,
                    computable_loss=computable_loss,
                    initializer=initializer,
                    not_init_flag=True,
                    hybridize_flag=hybridize_flag,
                    dropout_rate=dropout_rate
                )
            encoder.embedding_weignt = positional_embedding_weignt
        else:
            if isinstance(encoder, TransformerModel) is False:
                raise TypeError("The type of `encoder` must be `TransformerModel`.")

        if decoder is None:
            if hidden_dim is None or hidden_dim == depth_dim:
                decoder = TransformerDecoder(
                    head_n=head_n,
                    depth_dim=depth_dim,
                    layer_n=layer_n,
                    self_attention_activation_list=self_attention_activation_list,
                    multi_head_attention_activation_list=multi_head_attention_activation_list,
                    fc_activation_list=fc_activation_list,
                    computable_loss=computable_loss,
                    initializer=initializer,
                    not_init_flag=True,
                    hybridize_flag=hybridize_flag
                )
            else:
                decoder = TransformerDecoder(
                    head_n=head_n,
                    depth_dim=hidden_dim,
                    output_dim=depth_dim,
                    layer_n=layer_n,
                    self_attention_activation_list=self_attention_activation_list,
                    multi_head_attention_activation_list=multi_head_attention_activation_list,
                    fc_activation_list=fc_activation_list,
                    computable_loss=computable_loss,
                    initializer=initializer,
                    not_init_flag=True,
                    hybridize_flag=hybridize_flag
                )
            decoder.embedding_weignt = positional_embedding_weignt
        else:
            if isinstance(decoder, TransformerModel) is False:
                raise TypeError("The type of `decoder` must be `TransformerModel`.")

        logger = getLogger("accelbrainbase")
        self.logger = logger

        if initializer is None:
            self.initializer = mx.initializer.Xavier(
                rnd_type="gaussian", 
                factor_type="in", 
                magnitude=2
            )
        else:
            if isinstance(initializer, mx.initializer.Initializer) is False:
                raise TypeError("The type of `initializer` must be `mxnet.initializer.Initializer`.")
            self.initializer = initializer

        self.encoder = encoder
        self.decoder = decoder

        with self.name_scope():
            if hidden_dim is not None and hidden_dim != depth_dim:
                self.encoder_hidden_fc = gluon.nn.Dense(
                    hidden_dim, 
                    use_bias=True,
                    flatten=False,
                )
                self.register_child(self.encoder_hidden_fc)
                self.decoder_hidden_fc = gluon.nn.Dense(
                    hidden_dim, 
                    use_bias=True,
                    flatten=False,
                )
                self.register_child(self.decoder_hidden_fc)
            else:
                self.encoder_hidden_fc = None
                self.decoder_hidden_fc = None

            self.register_child(self.encoder)
            self.register_child(self.decoder)

        if self.init_deferred_flag is False:
            try:
                self.collect_params().initialize(self.initializer, force_reinit=True, ctx=ctx)
                self.trainer = gluon.Trainer(
                    self.collect_params(), 
                    optimizer_name, 
                    {
                        "learning_rate": learning_rate,
                        "wd": weight_decay
                    }
                )
                if hybridize_flag is True:
                    self.encoder.hybridize()
                    self.decoder.hybridize()

            except InitDeferredError:
                self.logger.debug("The initialization should be deferred.")

        for v in regularizatable_data_list:
            if isinstance(v, RegularizatableData) is False:
                raise TypeError("The type of values of `regularizatable_data_list` must be `RegularizatableData`.")
        self.__regularizatable_data_list = regularizatable_data_list

        self.__learning_rate = learning_rate
        self.__learning_attenuate_rate = learning_attenuate_rate
        self.__attenuate_epoch = attenuate_epoch

        self.seq_len = seq_len

    def collect_params(self, select=None):
        '''
        Overrided `collect_params` in `mxnet.gluon.HybridBlok`.
        '''
        params_dict = self.encoder.collect_params(select)
        params_dict.update(self.decoder.collect_params(select))
        if self.encoder_hidden_fc is not None:
            params_dict.update(self.encoder_hidden_fc.collect_params(select))
        if self.decoder_hidden_fc is not None:
            params_dict.update(self.decoder_hidden_fc.collect_params(select))
        return params_dict

    def learn(self, iteratable_data):
        '''
        Learn samples drawn by `IteratableData.generate_learned_samples()`.

        Args:
            iteratable_data:     is-a `TransformerIterator`.
        '''
        if isinstance(iteratable_data, TransformerIterator) is False:
            raise TypeError("The type of `iteratable_data` must be `TransformerIterator`.")

        self.__loss_list = []
        learning_rate = self.__learning_rate

        try:
            epoch = 0
            iter_n = 0
            for encoded_observed_arr, decoded_observed_arr, encoded_mask_arr, decoded_mask_arr, test_encoded_observed_arr, test_decoded_observed_arr, test_encoded_mask_arr, test_decoded_mask_arr, training_target_arr, test_target_arr in iteratable_data.generate_learned_samples():
                self.epoch = epoch
                if ((epoch + 1) % self.__attenuate_epoch == 0):
                    learning_rate = learning_rate * self.__learning_attenuate_rate
                    self.trainer.set_learning_rate(learning_rate)

                with autograd.record():
                    pred_arr = self.inference(
                        encoded_observed_arr, 
                        decoded_observed_arr, 
                        encoded_mask_arr, 
                        decoded_mask_arr
                    )
                    loss = self.compute_loss(
                        pred_arr,
                        training_target_arr
                    )
                loss.backward()
                self.trainer.step(encoded_observed_arr.shape[0])
                self.regularize()

                if (iter_n+1) % int(iteratable_data.iter_n / iteratable_data.epochs) == 0:
                    test_pred_arr = self.inference(
                        test_encoded_observed_arr, 
                        test_decoded_observed_arr, 
                        test_encoded_mask_arr, 
                        test_decoded_mask_arr
                    )

                    test_loss = self.compute_loss(
                        test_pred_arr,
                        test_target_arr
                    )
                    self.__loss_list.append((loss.asnumpy().mean(), test_loss.asnumpy().mean()))
                    self.logger.debug("Epochs: " + str(epoch + 1) + " Train loss: " + str(loss.asnumpy().mean()) + " Test loss: " + str(test_loss.asnumpy().mean()))
                    epoch += 1
                iter_n += 1

        except KeyboardInterrupt:
            self.logger.debug("Interrupt.")

        self.logger.debug("end. ")

    def inference(
        self, 
        encoded_observed_arr, 
        decoded_observed_arr, 
        encoded_mask_arr=None,
        decoded_mask_arr=None,
    ):
        '''
        Inference samples drawn by `IteratableData.generate_inferenced_samples()`.

        Args:
            encoded_observed_arr:   rank-3 Array like or sparse matrix as the observed data points.
                                    The shape is: (batch size, the length of sequence, feature points)

            decoded_observed_arr:   rank-3 Array like or sparse matrix as the observed data points.
                                    The shape is: (batch size, the length of sequence, feature points)

            encoded_mask_arr:       rank-3 Array like or sparse matrix as the observed data points.
                                    The shape is: (batch size, the length of sequence, feature points)

            decoded_mask_arr:       rank-3 Array like or sparse matrix as the observed data points.
                                    The shape is: (batch size, the length of sequence, feature points)

        Returns:
            `mxnet.ndarray` of inferenced feature points.
        '''
        if encoded_mask_arr is None:
            encoded_mask_arr = nd.ones(
                shape=(encoded_observed_arr.shape[0], 1, 1, 1), 
                ctx=encoded_observed_arr.context
            )
        if decoded_mask_arr is None:
            decoded_mask_arr = nd.ones(
                shape=(decoded_observed_arr.shape[0], 1, 1, 1), 
                ctx=decoded_observed_arr.context
            )

        return self(
            encoded_observed_arr, 
            decoded_observed_arr, 
            encoded_mask_arr,
            decoded_mask_arr,
        )

    def compute_loss(self, pred_arr, labeled_arr):
        '''
        Compute loss.

        Args:
            pred_arr:       `mxnet.ndarray` or `mxnet.symbol`.
            labeled_arr:    `mxnet.ndarray` or `mxnet.symbol`.

        Returns:
            loss.
        '''
        return self.__computable_loss(pred_arr, labeled_arr)

    def hybrid_forward(
        self, 
        F, 
        encoded_observed_arr, 
        decoded_observed_arr, 
        encoded_mask_arr=None,
        decoded_mask_arr=None,
    ):
        '''
        Hybrid forward with Gluon API.

        Args:
            F:                      `mxnet.ndarray` or `mxnet.symbol`.
            encoded_observed_arr:   rank-3 Array like or sparse matrix as the observed data points.
                                    The shape is: (batch size, the length of sequence, feature points)

            decoded_observed_arr:   rank-3 Array like or sparse matrix as the observed data points.
                                    The shape is: (batch size, the length of sequence, feature points)

            encoded_mask_arr:       rank-3 Array like or sparse matrix as the observed data points.
                                    The shape is: (batch size, the length of sequence, feature points)

            decoded_mask_arr:       rank-3 Array like or sparse matrix as the observed data points.
                                    The shape is: (batch size, the length of sequence, feature points)
        
        Returns:
            `mxnet.ndarray` or `mxnet.symbol` of inferenced feature points.
        '''
        # rank-3
        return self.forward_propagation(
            F, 
            encoded_observed_arr, 
            decoded_observed_arr, 
            encoded_mask_arr,
            decoded_mask_arr,
        )

    def forward_propagation(
        self, 
        F, 
        encoded_observed_arr, 
        decoded_observed_arr, 
        encoded_mask_arr=None,
        decoded_mask_arr=None,
    ):
        '''
        Hybrid forward with Gluon API.

        Args:
            F:                      `mxnet.ndarray` or `mxnet.symbol`.
            encoded_observed_arr:   rank-3 Array like or sparse matrix as the observed data points.
                                    The shape is: (batch size, the length of sequence, feature points)

            decoded_observed_arr:   rank-3 Array like or sparse matrix as the observed data points.
                                    The shape is: (batch size, the length of sequence, feature points)

            encoded_mask_arr:       rank-3 Array like or sparse matrix as the observed data points.
                                    The shape is: (batch size, the length of sequence, feature points)

            decoded_mask_arr:       rank-3 Array like or sparse matrix as the observed data points.
                                    The shape is: (batch size, the length of sequence, feature points)
        
        Returns:
            `mxnet.ndarray` or `mxnet.symbol` of inferenced feature points.
        '''
        if self.encoder_hidden_fc is not None:
            encoded_observed_arr = self.encoder_hidden_fc(encoded_observed_arr)
        if self.decoder_hidden_fc is not None:
            decoded_observed_arr = self.decoder_hidden_fc(decoded_observed_arr)

        steps_arr = F.arange(self.seq_len)
        mask_arr = F.broadcast_lesser_equal(
            steps_arr.reshape((1, -1)),
            steps_arr.reshape((-1, 1))
        )
        ones_arr = F.ones_like(steps_arr)
        seq_len_arr = ones_arr * self.seq_len
        batch_mask_arr = F.broadcast_lesser(
            steps_arr.reshape((1, -1)),
            seq_len_arr.reshape((-1, 1))
        )
        _decoded_mask_arr = F.broadcast_mul(batch_mask_arr, F.expand_dims(mask_arr, 0))
        _decoded_mask_arr = F.expand_dims(_decoded_mask_arr, 0)
        _decoded_mask_arr = _decoded_mask_arr + 1e-08

        if decoded_mask_arr is None:
            decoded_mask_arr = _decoded_mask_arr
        else:
            decoded_mask_arr = F.broadcast_add(decoded_mask_arr, _decoded_mask_arr)

        encoded_arr = self.encoder.inference(
            encoded_observed_arr,
            encoded_mask_arr
        )
        self.feature_points_arr = encoded_arr
        decoded_arr = self.decoder.inference(
            decoded_observed_arr, 
            encoded_arr, 
            decoded_mask_arr,
            encoded_mask_arr,
        )
        return decoded_arr

    def extract_learned_dict(self):
        '''
        Extract (pre-) learned parameters.

        Returns:
            `dict` of the parameters.
        '''
        params_dict = self.collect_params()
        
        params_arr_dict = {}
        for k in params_dict:
            params_arr_dict.setdefault(k, params_dict[k].data())

        return params_arr_dict

    def regularize(self):
        '''
        Regularization.
        '''
        params_dict = self.extract_learned_dict()
        for regularizatable in self.__regularizatable_data_list:
            params_dict = regularizatable.regularize(params_dict)

        for k, params in self.collect_params().items():
            params.set_data(params_dict[k])

    def __rename_file(self, filename):
        filename_list = filename.split(".")
        _format = filename_list[-1]
        g_filename = filename.replace("." + _format, "_encoder." + _format)
        d_filename = filename.replace("." + _format, "_decoder." + _format)
        return g_filename, d_filename

    def save_parameters(self, filename):
        '''
        Save parameters to files.

        Args:
            filename:       File name.
        '''
        e_filename, d_filename = self.__rename_file(filename)
        self.encoder.save_parameters(e_filename)
        self.decoder.save_parameters(d_filename)

    def load_parameters(self, filename, ctx=None, allow_missing=False, ignore_extra=False):
        '''
        Load parameters to files.

        Args:
            filename:       File name.
            ctx:            `mx.cpu()` or `mx.gpu()`.
            allow_missing:  `bool` of whether to silently skip loading parameters not represents in the file.
            ignore_extra:   `bool` of whether to silently ignre parameters from the file that are not present in this `Block`.
        '''
        e_filename, d_filename = self.__rename_file(filename)
        self.encoder.load_parameters(e_filename, ctx=ctx, allow_missing=allow_missing, ignore_extra=ignore_extra)
        self.decoder.load_parameters(d_filename, ctx=ctx, allow_missing=allow_missing, ignore_extra=ignore_extra)

    def set_readonly(self, value):
        ''' setter '''
        raise TypeError("This property must be read-only.")

    def get_init_deferred_flag(self):
        ''' getter for `bool` that means initialization in this class will be deferred or not.'''
        return self.__init_deferred_flag
    
    def set_init_deferred_flag(self, value):
        ''' setter for `bool` that means initialization in this class will be deferred or not.'''
        self.__init_deferred_flag = value

    init_deferred_flag = property(get_init_deferred_flag, set_init_deferred_flag)

    def get_loss_arr(self):
        ''' getter '''
        return np.array(self.__loss_list)
    
    def set_loss_arr(self, value):
        ''' setter '''
        raise TypeError("This property must be read-only.")
    
    loss_arr = property(get_loss_arr, set_loss_arr)



# -*- coding: utf-8 -*-
from accelbrainbase.noiseable_data import NoiseableData
import numpy as np


class GaussNoise(NoiseableData):
    '''
    Gauss noise function.
    '''

    def __init__(self, mu=0.0, sigma=1.0):
        '''
        Init.

        Args:
            mu:     Mean of the Gauss distribution.
            sigma:  Standard deviation of the Gauss distribution.
        '''
        self.__mu = mu
        self.__sigma = sigma

    def noise(self, arr):
        '''
        Noise.

        Args:
            F:      `mx.ndarray` or `mx.symbol`.
            arr:    `mx.nd.array` or `mx.sym.array`.
        
        Returns:
            `mx.nd.array` or `mx.sym.array`.
        '''
        return arr + npz.random.normal(
            size=arr.shape, 
            loc=self.__mu, 
            scale=self.__sigma
        )



# -*- coding: utf-8 -*-
from accelbrainbase.observabledata._torch.convolutionalneuralnetworks.convolutional_auto_encoder import ConvolutionalAutoEncoder
from accelbrainbase.noiseabledata._torch.gauss_noise import GaussNoise
from accelbrainbase.noiseable_data import NoiseableData

import torch
import torch.nn as nn
import torch.nn.functional as F
from torch.optim.adam import Adam
import numpy as np
from logging import getLogger


class ConvolutionalLadderNetworks(ConvolutionalAutoEncoder):
    '''
    Ladder Networks with a Stacked convolutional Auto-Encoder.

    In most classification problems, finding and producing labels for the samples is hard. 
    In many cases plenty of unlabeled data existand it seems obvious that using them should 
    improve the results. For instance, there are plenty of unlabeled images available and 
    in most image classification tasks there are vastly more bits of information in the 
    statistical structure of input images than in their labels.

    It is argued here that the reason why unsupervised learning has not been able to improve 
    results is that most current versions are incompatible with supervised learning. 
    The problem is that many un-supervised learning methods try to represent as much 
    information about the original data as possible whereas supervised learning tries to 
    filter out all the information which is irrelevant for the task at hand.

    Ladder network is an Auto-Encoder which can discard information.
    Unsupervised learning needs to toleratediscarding information in order to 
    work well with supervised learning. Many unsupervised learning methods are 
    not good at this but one class of models stands out as an exception: 
    hierarchical latent variable models. Unfortunately their derivation can be 
    quite complicated and often involves approximations which compromise their per-formance.
    
    A simpler alternative is offered by Auto-Encoders which also have the benefit 
    of being compatible with standard supervised feedforward networks. They would be a 
    promising candidate for combining supervised and unsupervised learning but unfortunately 
    Auto-Encoders normally correspond to latent variable models with a single layer of 
    stochastic variables, that is, they do not tolerate discarding information.

    Ladder network makes it possible to solve that problem by settting recursive derivation 
    of the learning rule with a distributed cost function, building denoisng Auto-Encoder recursively.
    Normally denoising Auto-Encoders have a fixed input but the cost functions on the higher layers 
    can influence their input mappings and this creates a bias towards PCA-type solutions.

    References:
        - Bengio, Y., Lamblin, P., Popovici, D., & Larochelle, H. (2007). Greedy layer-wise training of deep networks. In Advances in neural information processing systems (pp. 153-160).
        - Dumoulin, V., & V,kisin, F. (2016). A guide to convolution arithmetic for deep learning. arXiv preprint arXiv:1603.07285.
        - Erhan, D., Bengio, Y., Courville, A., Manzagol, P. A., Vincent, P., & Bengio, S. (2010). Why does unsupervised pre-training help deep learning?. Journal of Machine Learning Research, 11(Feb), 625-660.
        - Erhan, D., Courville, A., & Bengio, Y. (2010). Understanding representations learned in deep architectures. Department dInformatique et Recherche Operationnelle, University of Montreal, QC, Canada, Tech. Rep, 1355, 1.
        - Goodfellow, I., Bengio, Y., & Courville, A. (2016). Deep learning (adaptive computation and machine learning series). Adaptive Computation and Machine Learning series, 800.
        - Masci, J., Meier, U., Cireşan, D., & Schmidhuber, J. (2011, June). Stacked convolutional auto-encoders for hierarchical feature extraction. In International Conference on Artificial Neural Networks (pp. 52-59). Springer, Berlin, Heidelberg.
        - Rasmus, A., Berglund, M., Honkala, M., Valpola, H., & Raiko, T. (2015). Semi-supervised learning with ladder networks. In Advances in neural information processing systems (pp. 3546-3554).
        - Valpola, H. (2015). From neural PCA to deep unsupervised learning. In Advances in Independent Component Analysis and Learning Machines (pp. 143-171). Academic Press.
    '''

    # is-a `NoiseableData`.
    __noiseable_data = GaussNoise(mu=0.0, sigma=1e-03)

    def get_noiseable_data(self):
        ''' getter '''
        return self.__noiseable_data

    def set_noiseable_data(self, value):
        ''' setter '''
        if isinstance(value, NoiseableData) is True:
            self.__noiseable_data = value
        else:
            raise TypeError("The type of `noiseable_data` must be `NoiseableData`.")

    noiseable_data = property(get_noiseable_data, set_noiseable_data)

    __recoding_ld_loss = False

    def get_recoding_ld_loss(self):
        ''' getter '''
        return self.__recoding_ld_loss
    
    def set_recoding_ld_loss(self, value):
        ''' setter '''
        self.__recoding_ld_loss = value

    recoding_ld_loss = property(get_recoding_ld_loss, set_recoding_ld_loss)

    # alpha weight.
    __alpha = 1e-05
    # sigma weight.
    __sigma = 0.7
    # mu weight.
    __mu = 0.7

    def forward(self, x):
        '''
        Hybrid forward with Gluon API.

        Args:
            x:      `tensor` of observed data points.
        
        Returns:
            `tensor` of inferenced feature points.
        '''
        noised_x = self.noiseable_data.noise(x)
        for i in range(len(self.encoder.hidden_units_list)):
            x = self.encoder.hidden_units_list[i](x)
            noised_x = self.encoder.hidden_units_list[i](noised_x)
            if self.encoder.hidden_activation_list[i] == "identity_adjusted":
                x = x / torch.sum(torch.ones_like(x))
                noised_x = noised_x / torch.sum(torch.ones_like(noised_x))
            elif self.encoder.hidden_activation_list[i] != "identity":
                x = self.encoder.hidden_activation_list[i](x)
                noised_x = self.encoder.hidden_activation_list[i](noised_x)
            if self.encoder.hidden_dropout_rate_list[i] is not None:
                x = self.encoder.hidden_dropout_rate_list[i](x)
                noised_x = self.encoder.hidden_dropout_rate_list[i](noised_x)

            if self.encoder.hidden_batch_norm_list[i] is not None:
                x = self.encoder.hidden_batch_norm_list[i](x)
                noised_x = self.encoder.hidden_batch_norm_list[i](noised_x)

            noised_x = self.noiseable_data.noise(noised_x)
            alpha_arr, sigma_arr, mu_arr = self.forward_ladder_net(x, noised_x)
            x = torch.add(x, (alpha_arr * self.alpha))
            x = torch.add(x, (sigma_arr * self.sigma))
            x = torch.add(x, (mu_arr * self.mu))

        if self.output_nn is None:
            self.feature_points_arr = x
        else:
            inner_x = self.output_nn(x)
            self.feature_points_arr = inner_x

        if self.recoding_ld_loss is True:
            if x.requires_grad is True:
                self.alpha_encoder_loss_list.append(torch.mean(torch.square(alpha_arr)).to("cpu").detach().numpy())
                self.sigma_encoder_loss_list.append(torch.mean(torch.square(sigma_arr)).to("cpu").detach().numpy())
                self.mu_encoder_loss_list.append(torch.mean(torch.square(mu_arr)).to("cpu").detach().numpy())
            else:
                self.alpha_encoder_test_loss_list.append(torch.mean(torch.square(alpha_arr)).to("cpu").detach().numpy())
                self.sigma_encoder_test_loss_list.append(torch.mean(torch.square(sigma_arr)).to("cpu").detach().numpy())
                self.mu_encoder_test_loss_list.append(torch.mean(torch.square(mu_arr)).to("cpu").detach().numpy())

        noised_x = self.noiseable_data.noise(x)
        for i in range(len(self.decoder.hidden_units_list)):
            x = self.decoder.hidden_units_list[i](x)
            noised_x = self.decoder.hidden_units_list[i](noised_x)
            if self.decoder.hidden_activation_list[i] == "identity_adjusted":
                x = x / torch.sum(torch.ones_like(x))
                noised_x = noised_x / torch.sum(torch.ones_like(noised_x))
            elif self.decoder.hidden_activation_list[i] != "identity":
                x = self.decoder.hidden_activation_list[i](x)
                noised_x = self.decoder.hidden_activation_list[i](noised_x)
            if self.decoder.hidden_dropout_rate_list[i] is not None:
                x = self.decoder.hidden_dropout_rate_list[i](x)
                noised_x = self.decoder.hidden_dropout_rate_list[i](noised_x)

            if self.decoder.hidden_batch_norm_list[i] is not None:
                x = self.decoder.hidden_batch_norm_list[i](x)
                noised_x = self.decoder.hidden_batch_norm_list[i](noised_x)

            noised_x = self.noiseable_data.noise(noised_x)
            alpha_arr, sigma_arr, mu_arr = self.forward_ladder_net(x, noised_x)

            x = torch.add(x, (alpha_arr * self.alpha))
            x = torch.add(x, (sigma_arr * self.sigma))
            x = torch.add(x, (mu_arr * self.mu))

        if self.decoder.output_nn is not None:
            x = self.decoder.output_nn(x)

        if self.recoding_ld_loss is True:
            if x.requires_grad is True:
                self.alpha_decoder_loss_list.append(torch.mean(torch.square(alpha_arr)).to("cpu").detach().numpy())
                self.sigma_decoder_loss_list.append(torch.mean(torch.square(sigma_arr)).to("cpu").detach().numpy())
                self.mu_decoder_loss_list.append(torch.mean(torch.square(mu_arr)).to("cpu").detach().numpy())
            else:
                self.alpha_decoder_test_loss_list.append(torch.mean(torch.square(alpha_arr)).to("cpu").detach().numpy())
                self.sigma_decoder_test_loss_list.append(torch.mean(torch.square(sigma_arr)).to("cpu").detach().numpy())
                self.mu_decoder_test_loss_list.append(torch.mean(torch.square(mu_arr)).to("cpu").detach().numpy())

        return x

    def forward_ladder_net(self, x, noised_x):
        pow_f = torch.pow

        hidden_arr = x.reshape(self.batch_size, -1)
        row_arr = torch.ones_like(hidden_arr)
        row = torch.mean(torch.sum(row_arr, axis=1))
        sigma_arr = torch.mm(
            hidden_arr,
            hidden_arr.T
        )
        sigma_arr = sigma_arr + 1e-08
        sigma_arr = torch.eye(self.batch_size, device=sigma_arr.device) - pow_f(sigma_arr, -1)
        sigma_arr = torch.mean(sigma_arr, axis=0)
        sigma_arr = torch.reshape(sigma_arr, shape=(
            self.batch_size,
            1,
            1,
            1
        ))
        sigma_arr = torch.div(sigma_arr, row)
        mu_arr = torch.div(torch.square(x), row)
        alpha_arr = x - noised_x

        return alpha_arr, sigma_arr, mu_arr

    def set_readonly(self, value):
        ''' setter '''
        raise TypeError("This property must be read-only.")

    def get_alpha(self):
        ''' getter '''
        return self.__alpha
    
    def set_alpha(self, value):
        ''' setter '''
        self.__alpha = value
    
    alpha = property(get_alpha, set_alpha)

    def get_sigma(self):
        ''' getter '''
        return self.__sigma
    
    def set_sigma(self, value):
        ''' setter '''
        self.__sigma = value
    
    sigma = property(get_sigma, set_sigma)

    def get_mu(self):
        ''' getter '''
        return self.__mu
    
    def set_mu(self, value):
        ''' setter '''
        self.__mu = value
    
    mu = property(get_mu, set_mu)

    def get_alpha_loss_arr(self):
        '''
        getter
        
        Returns:
            Logs of alpha losses. The shape is ...
            - encoder's train losses of alpha.
            - encoder's test losses of alpha.
            - decoder's train losses of alpha.
            - decoder's test losses of alpha.
        '''
        return np.c_[
            np.array(self.__alpha_encoder_loss_list),
            np.array(self.__alpha_encoder_test_loss_list[:len(self.__alpha_encoder_loss_list)]),
            np.array(self.__alpha_decoder_loss_list),
            np.array(self.__alpha_decoder_test_loss_list[:len(self.__alpha_decoder_loss_list)]),
        ]

    def get_sigma_loss_arr(self):
        '''
        getter
        
        Returns:
            Logs of sigma losses. The shape is ...
            - encoder's train losses of sigma.
            - encoder's test losses of sigma.
            - decoder's train losses of sigma.
            - decoder's test losses of sigma.
        '''
        return np.c_[
            np.array(self.__sigma_encoder_loss_list),
            np.array(self.__sigma_encoder_test_loss_list[:len(self.__sigma_encoder_loss_list)]),
            np.array(self.__sigma_decoder_loss_list),
            np.array(self.__sigma_decoder_test_loss_list[:len(self.__sigma_decoder_loss_list)])
        ]

    def get_mu_loss_arr(self):
        '''
        getter
        
        Returns:
            Logs of mu losses. The shape is ...
            - encoder's train losses of mu.
            - encoder's test losses of mu.
            - decoder's train losses of mu.
            - decoder's test losses of mu.
        '''
        return np.c_[
            np.array(self.__mu_encoder_loss_list),
            np.array(self.__mu_encoder_test_loss_list[:len(self.__mu_encoder_loss_list)]),
            np.array(self.__mu_decoder_loss_list),
            np.array(self.__mu_decoder_test_loss_list[:len(self.__mu_decoder_loss_list)])
        ]

    alpha_loss_arr = property(get_alpha_loss_arr, set_readonly)
    sigma_loss_arr = property(get_sigma_loss_arr, set_readonly)
    mu_loss_arr = property(get_mu_loss_arr, set_readonly)

    # `list` of encoder's alpha loss.
    __alpha_encoder_loss_list = []

    def get_alpha_encoder_loss_list(self):
        ''' getter '''
        return self.__alpha_encoder_loss_list
    
    def set_alpha_encoder_loss_list(self, value):
        ''' setter '''
        self.__alpha_encoder_loss_list = value

    alpha_encoder_loss_list = property(get_alpha_encoder_loss_list, set_alpha_encoder_loss_list)

    # `list` of encoder's sigma loss.
    __sigma_encoder_loss_list = []

    def get_sigma_encoder_loss_list(self):
        ''' getter '''
        return self.__sigma_encoder_loss_list
    
    def set_sigma_encoder_loss_list(self, value):
        ''' setter '''
        self.__sigma_encoder_loss_list = value

    sigma_encoder_loss_list = property(get_sigma_encoder_loss_list, set_sigma_encoder_loss_list)

    # `list` of encoder's mu loss.
    __mu_encoder_loss_list = []

    def get_mu_encoder_loss_list(self):
        ''' getter '''
        return self.__mu_encoder_loss_list
    
    def set_mu_encoder_loss_list(self, value):
        ''' setter '''
        self.__mu_encoder_loss_list = value
    
    mu_encoder_loss_list = property(get_mu_encoder_loss_list, set_mu_encoder_loss_list)

    # `list` of decoder's alpha loss.
    __alpha_decoder_loss_list = []

    def get_alpha_decoder_loss_list(self):
        ''' getter '''
        return self.__alpha_decoder_loss_list
    
    def set_alpha_decoder_loss_list(self, value):
        ''' setter '''
        self.__alpha_decoder_loss_list = value
    
    alpha_decoder_loss_list = property(get_alpha_decoder_loss_list, set_alpha_decoder_loss_list)

    # `list` of decoder's sigma loss.
    __sigma_decoder_loss_list = []

    def get_sigma_decoder_loss_list(self):
        ''' getter '''
        return self.__sigma_decoder_loss_list
    
    def set_sigma_decoder_loss_list(self, value):
        ''' setter '''
        self.__sigma_decoder_loss_list = value
    
    sigma_decoder_loss_list = property(get_sigma_decoder_loss_list, set_sigma_decoder_loss_list)

    # `list` of decoder's mu loss.
    __mu_decoder_loss_list = []

    def get_mu_decoder_loss_list(self):
        ''' getter '''
        return self.__mu_decoder_loss_list

    def set_mu_decoder_loss_list(self, value):
        ''' setter '''
        self.__mu_decoder_loss_list = value

    mu_decoder_loss_list = property(get_mu_decoder_loss_list, set_mu_decoder_loss_list)

    # `list` of encoder's alpha test loss.
    __alpha_encoder_test_loss_list = []

    def get_alpha_encoder_test_loss_list(self):
        ''' getter '''
        return self.__alpha_encoder_test_loss_list

    def set_alpha_encoder_test_loss_list(self, value):
        ''' setter '''
        self.__alpha_encoder_test_loss_list = value

    alpha_encoder_test_loss_list = property(get_alpha_encoder_test_loss_list, set_alpha_encoder_test_loss_list)

    # `list` of encoder's sigma test loss.
    __sigma_encoder_test_loss_list = []

    def get_sigma_encoder_test_loss_list(self):
        ''' getter '''
        return self.__sigma_encoder_test_loss_list
    
    def set_sigma_encoder_test_loss_list(self, value):
        ''' setter '''
        self.__sigma_encoder_test_loss_list = value

    sigma_encoder_test_loss_list = property(get_sigma_encoder_test_loss_list, set_sigma_encoder_test_loss_list)

    # `list` of encoder's mu test loss.
    __mu_encoder_test_loss_list = []

    def get_mu_encoder_test_loss_list(self):
        ''' getter '''
        return self.__mu_encoder_test_loss_list

    def set_mu_encoder_test_loss_list(self, value):
        ''' setter '''
        self.__mu_encoder_test_loss_list = value

    mu_encoder_test_loss_list = property(get_mu_encoder_test_loss_list, set_mu_encoder_test_loss_list)

    # `list` of decoder's alpha test loss.
    __alpha_decoder_test_loss_list = []

    def get_alpha_decoder_test_loss_list(self):
        ''' getter '''
        return self.__alpha_decoder_test_loss_list

    def set_alpha_decoder_test_loss_list(self, value):
        ''' setter '''
        self.__alpha_decoder_test_loss_list = value

    alpha_decoder_test_loss_list = property(get_alpha_decoder_test_loss_list, set_alpha_decoder_test_loss_list)

    # `list` of decoder's sigma test loss.
    __sigma_decoder_test_loss_list = []

    def get_sigma_decoder_test_loss_list(self):
        ''' getter '''
        return self.__sigma_decoder_test_loss_list

    def set_sigma_decoder_test_loss_list(self, value):
        ''' setter '''
        self.__sigma_decoder_test_loss_list = value

    sigma_decoder_test_loss_list = property(get_sigma_decoder_test_loss_list, set_sigma_decoder_test_loss_list)

    # `list` of decoder's mu test loss.
    __mu_decoder_test_loss_list = []

    def get_mu_decoder_test_loss_list(self):
        ''' getter '''
        return self.__mu_decoder_test_loss_list
    
    def set_mu_decoder_test_loss_list(self, value):
        ''' setter '''
        self.__mu_decoder_test_loss_list = value
    
    mu_decoder_test_loss_list = property(get_mu_decoder_test_loss_list, set_mu_decoder_test_loss_list)



# -*- coding: utf-8 -*-
from accelbrainbase.observabledata._torch.convolutionalneuralnetworks.convolutional_auto_encoder import ConvolutionalAutoEncoder
import numpy as np
from logging import getLogger


class ContractiveCAE(ConvolutionalAutoEncoder):
    '''
    Convolutional Contracitve Auto-Encoder.

    A stack of Convolutional Auto-Encoder (Masci, J., et al., 2011) 
    forms a convolutional neural network(CNN), which are among the most successful models 
    for supervised image classification.  Each Convolutional Auto-Encoder is trained 
    using conventional on-line gradient descent without additional regularization terms.
    
    In this library, Convolutional Auto-Encoder is also based on Encoder/Decoder scheme.
    The encoder is to the decoder what the Convolution is to the Deconvolution.
    The Deconvolution also called transposed convolutions 
    "work by swapping the forward and backward passes of a convolution." (Dumoulin, V., & Visin, F. 2016, p20.)

    The First-Order Contractive Auto-Encoder(Rifai, S., et al., 2011) executes 
    the representation learning by adding a penalty term to the classical 
    reconstruction cost function. This penalty term corresponds to 
    the Frobenius norm of the Jacobian matrix of the encoder activations
    with respect to the input and results in a localized space 
    contraction which in turn yields robust features on the activation layer.

    Analogically, the Contractive Convolutional Auto-Encoder calculates the penalty term.
    But it differs in that the operation of the deconvolution intervenes insted of inner product.

    References:
        - Dumoulin, V., & V,kisin, F. (2016). A guide to convolution arithmetic for deep learning. arXiv preprint arXiv:1603.07285.
        - Kamyshanska, H., & Memisevic, R. (2014). The potential energy of an autoencoder. IEEE transactions on pattern analysis and machine intelligence, 37(6), 1261-1273.
        - Masci, J., Meier, U., Cireşan, D., & Schmidhuber, J. (2011, June). Stacked convolutional auto-encoders for hierarchical feature extraction. In International Conference on Artificial Neural Networks (pp. 52-59). Springer, Berlin, Heidelberg.
        - Rifai, S., Vincent, P., Muller, X., Glorot, X., & Bengio, Y. (2011, June). Contractive auto-encoders: Explicit invariance during feature extraction. In Proceedings of the 28th International Conference on International Conference on Machine Learning (pp. 833-840). Omnipress.
        - Rifai, S., Mesnil, G., Vincent, P., Muller, X., Bengio, Y., Dauphin, Y., & Glorot, X. (2011, September). Higher order contractive auto-encoder. In Joint European Conference on Machine Learning and Knowledge Discovery in Databases (pp. 645-660). Springer, Berlin, Heidelberg.
    '''

    # penalty lambda.
    __penalty_lambda = 1e-05

    def forward(self, x):
        '''
        forward with Gluon API.

        Args:
            x:      `mxnet.ndarray` of observed data points.
        
        Returns:
            `mxnet.ndarray` or `mxnet.symbol` of inferenced feature points.
        '''
        decoded_arr = super().forward(x)
        penalty_arr = self.feature_points_arr * (1 - self.feature_points_arr)

        if self.penalty_lambda > 0:
            penalty_arr = self.decoder.forward(penalty_arr)
        
        return decoded_arr + (self.penalty_lambda * penalty_arr)

    def get_penalty_lambda(self):
        ''' getter for lambda. '''
        return self.__penalty_lambda
    
    def set_penalty_lambda(self, value):
        ''' setter for lambda. '''
        self.__penalty_lambda = value
    
    penalty_lambda = property(get_penalty_lambda, set_penalty_lambda)



# -*- coding: utf-8 -*-
from accelbrainbase.observabledata._torch.function_approximator import FunctionApproximator as _FunctionApproximator
import numpy as np
from logging import getLogger
import torch
from torch import nn
from torch.optim.adam import Adam


class FunctionApproximator(nn.Module, _FunctionApproximator):
    '''
    The function approximator for the Deep Q-Learning.

    The convolutional neural networks(CNNs) are hierarchical models 
    whose convolutional layers alternate with subsampling layers, 
    reminiscent of simple and complex cells in the primary visual cortex.
    
    Mainly, this class demonstrates that a CNNs can solve generalisation problems to learn 
    successful control policies from observed data points in complex 
    Reinforcement Learning environments. The network is trained with a variant of 
    the Q-learning algorithm, with stochastic gradient descent to update the weights.

    But there is no need for the function approximator to be a CNNs.
    We probide this interface that implements various models as 
    function approximations, not limited to CNNs.

    References:
        - Dumoulin, V., & V,kisin, F. (2016). A guide to convolution arithmetic for deep learning. arXiv preprint arXiv:1603.07285.
        - Masci, J., Meier, U., Cireşan, D., & Schmidhuber, J. (2011, June). Stacked convolutional auto-encoders for hierarchical feature extraction. In International Conference on Artificial Neural Networks (pp. 52-59). Springer, Berlin, Heidelberg.
        - Mnih, V., Kavukcuoglu, K., Silver, D., Graves, A., Antonoglou, I., Wierstra, D., & Riedmiller, M. (2013). Playing atari with deep reinforcement learning. arXiv preprint arXiv:1312.5602.

    '''

    def __init__(
        self, 
        model, 
        learning_rate=1e-05,
        ctx="cpu", 
        not_init_flag=False,
    ):
        '''
        Init.

        Args:
            model:                          is-a `mxnet.gluon.hybrid.hybridblock.HybridBlock`.
            initializer:                    is-a `mxnet.initializer` for parameters of model. If `None`, it is drawing from the Xavier distribution.
            learning_rate:                  `float` of learning rate.
            optimizer_name:                 `str` of name of optimizer.
            hybridize_flag:                  Call `mxnet.gluon.HybridBlock.hybridize()` or not.
            scale:                          `float` of scaling factor for initial parameters.
            ctx:                            `mx.cpu()` or `mx.gpu()`.

        '''
        super(FunctionApproximator, self).__init__()

        self.model = model
        logger = getLogger("accelbrainbase")
        self.__logger = logger

        self.__not_init_flag = not_init_flag

    def inference(self, observed_arr):
        '''
        Draw samples from the fake distribution.

        Args:
            observed_arr:       `mxnet.ndarray` or `mxnet.symbol` of observed data points.

        Returns:
            `Tuple` of `mxnet.ndarray`s.
        '''
        return self(observed_arr)

    def forward(self, x):
        '''
        Hybrid forward with Gluon API.

        Args:
            F:      `mxnet.ndarray` or `mxnet.symbol`.
            x:      `mxnet.ndarray` of observed data points.
        
        Returns:
            `mxnet.ndarray` or `mxnet.symbol` of inferenced feature points.
        '''
        return self.model(x)



# -*- coding: utf-8 -*-
from accelbrainbase.samplabledata.true_sampler import TrueSampler
from accelbrainbase.samplable_data import SamplableData
from accelbrainbase.iteratabledata.labeled_image_iterator import LabeledImageIterator


class LabeledTrueSampler(TrueSampler):
    '''
    The class to draw true labeled samples from distributions.

    References:
        - Bousmalis, K., Silberman, N., Dohan, D., Erhan, D., & Krishnan, D. (2017). Unsupervised pixel-level domain adaptation with generative adversarial networks. In Proceedings of the IEEE conference on computer vision and pattern recognition (pp. 3722-3731).
        - Goodfellow, I., Pouget-Abadie, J., Mirza, M., Xu, B., Warde-Farley, D., Ozair, S., ... & Bengio, Y. (2014). Generative adversarial nets. In Advances in neural information processing systems (pp. 2672-2680).
        - Mirza, M., & Osindero, S. (2014). Conditional generative adversarial nets. arXiv preprint arXiv:1411.1784.

    '''
    # is-a `ImageIterator`.
    __labeled_image_iterator = None

    def get_labeled_image_iterator(self):
        ''' getter '''
        if isinstance(self.__labeled_image_iterator, LabeledImageIterator) is False:
            raise TypeError("The type of `__labeled_image_iterator` must be `LabeledImageIterator`.")
        return self.__labeled_image_iterator

    def set_labeled_image_iterator(self, value):
        ''' setter '''
        if isinstance(value, LabeledImageIterator) is False:
            raise TypeError("The type of `__labeled_image_iterator` must be `LabeledImageIterator`.")
        self.__labeled_image_iterator = value
    
    labeled_image_iterator = property(get_labeled_image_iterator, set_labeled_image_iterator)

    def draw(self):
        '''
        Draw samples from distribtions.
        
        Returns:
            `Tuple` of `mx.nd.array`s.
        '''
        observed_arr = None
        label_arr = None
        for arr_tuple in self.labeled_image_iterator.generate_learned_samples():
            observed_arr = arr_tuple[0]
            label_arr = arr_tuple[1]
            break
        return observed_arr, label_arr



# -*- coding: utf-8 -*-
from accelbrainbase.iteratable_data import IteratableData
from accelbrainbase.regularizatable_data import RegularizatableData
from accelbrainbase.computable_loss import ComputableLoss
from accelbrainbase.observabledata._torch.convolutional_neural_networks import ConvolutionalNeuralNetworks

import torch
import torch.nn as nn
import torch.nn.functional as F
from torch.optim.adam import Adam

import numpy as np
from logging import getLogger


class ConvolutionalAutoEncoder(ConvolutionalNeuralNetworks):
    '''
    Convolutional Auto-Encoder.

    A stack of Convolutional Auto-Encoder (Masci, J., et al., 2011) 
    forms a convolutional neural network(CNN), which are among the most successful models 
    for supervised image classification.  Each Convolutional Auto-Encoder is trained 
    using conventional on-line gradient descent without additional regularization terms.
    
    In this library, Convolutional Auto-Encoder is also based on Encoder/Decoder scheme.
    The encoder is to the decoder what the Convolution is to the Deconvolution.
    The Deconvolution also called transposed convolutions 
    "work by swapping the forward and backward passes of a convolution." (Dumoulin, V., & Visin, F. 2016, p20.)

    References:
        - Dumoulin, V., & V,kisin, F. (2016). A guide to convolution arithmetic for deep learning. arXiv preprint arXiv:1603.07285.
        - Masci, J., Meier, U., Cireşan, D., & Schmidhuber, J. (2011, June). Stacked convolutional auto-encoders for hierarchical feature extraction. In International Conference on Artificial Neural Networks (pp. 52-59). Springer, Berlin, Heidelberg.
    '''

    # `bool` that means initialization in this class will be deferred or not.
    __init_deferred_flag = False

    def __init__(
        self,
        encoder,
        decoder,
        computable_loss,
        initializer_f=None,
        optimizer_f=None,
        encoder_optimizer_f=None,
        decoder_optimizer_f=None,
        learning_rate=1e-05,
        hidden_units_list=[],
        output_nn=None,
        hidden_dropout_rate_list=[],
        hidden_activation_list=[],
        hidden_batch_norm_list=[],
        ctx="cpu",
        regularizatable_data_list=[],
        scale=1.0,
        tied_weights_flag=True,
        init_deferred_flag=None,
        not_init_flag=False,
        wd=None,
    ):
        '''
        Init.

        Args:
            encoder:                        is-a `CNNHybrid`.
            decoder:                        is-a `CNNHybrid`.
            computable_loss:                is-a `ComputableLoss` or `mxnet.gluon.loss`.
            initializer_f:                  is-a `mxnet.initializer.Initializer` for parameters of model. If `None`, it is drawing from the Xavier distribution.
            learning_rate:                  `float` of learning rate.

            hidden_units_list:              `list` of `mxnet.gluon.nn._conv` in hidden layers.
            output_nn:                      is-a `NNHybrid` as output layers.
                                            If `None`, last layer in `hidden_units_list` will be considered as an output layer.

            hidden_dropout_rate_list:       `list` of `float` of dropout rate in hidden layers.

            optimizer_name:                 `str` of name of optimizer.

            hidden_activation_list:         `list` of act_type` in `mxnet.ndarray.Activation` or `mxnet.symbol.Activation` in input gate.
            hidden_batch_norm_list:         `list` of `mxnet.gluon.nn.BatchNorm` in hidden layers.

            ctx:                            `mx.cpu()` or `mx.gpu()`.
            regularizatable_data_list:      `list` of `RegularizatableData`.
            scale:                          `float` of scaling factor for initial parameters.
            tied_weights_flag:              `bool` of flag to tied weights or not.
            wd:                             `float` of parameter of weight decay.
            init_deferred_flag:             `bool` that means initialization in this class will be deferred or not.
        '''
        if isinstance(encoder, ConvolutionalNeuralNetworks) is False:
            raise TypeError("The type of `encoder` must be `ConvolutionalNeuralNetworks`.")
        if isinstance(decoder, ConvolutionalNeuralNetworks) is False:
            raise TypeError("The type of `decoder` must be `ConvolutionalNeuralNetworks`.")

        if len(hidden_units_list) != len(hidden_activation_list):
            raise ValueError("The length of `hidden_units_list` and `hidden_activation_list` must be equivalent.")

        if len(hidden_dropout_rate_list) != len(hidden_units_list):
            raise ValueError("The length of `hidden_dropout_rate_list` and `hidden_units_list` must be equivalent.")

        if isinstance(computable_loss, ComputableLoss) is False and isinstance(computable_loss, nn.modules.loss._Loss) is False:
            raise TypeError("The type of `computable_loss` must be `ComputableLoss` or `nn.modules.loss._Loss`.")

        logger = getLogger("accelbrainbase")
        self.__logger = logger

        if init_deferred_flag is None:
            init_deferred_flag = self.init_deferred_flag
        elif isinstance(init_deferred_flag, bool) is False:
            raise TypeError("The type of `init_deferred_flag` must be `bool`.")

        self.__not_init_flag = not_init_flag
        self.init_deferred_flag = True

        super().__init__(
            computable_loss=computable_loss,
            initializer_f=initializer_f,
            optimizer_f=optimizer_f,
            learning_rate=learning_rate,
            hidden_units_list=hidden_units_list,
            output_nn=output_nn,
            hidden_dropout_rate_list=hidden_dropout_rate_list,
            hidden_activation_list=hidden_activation_list,
            hidden_batch_norm_list=hidden_batch_norm_list,
            ctx=ctx,
            regularizatable_data_list=regularizatable_data_list,
            scale=scale,
        )
        self.init_deferred_flag = init_deferred_flag
        self.encoder = encoder
        self.decoder = decoder
        self.__tied_weights_flag = tied_weights_flag
        self.output_nn = output_nn
        self.optimizer_f = optimizer_f
        self.encoder_optimizer_f = encoder_optimizer_f
        self.decoder_optimizer_f = decoder_optimizer_f
        self.__computable_loss = computable_loss
        self.__learning_rate = learning_rate
        self.encoder_optimizer = None
        self.decoder_optimizer = None
        self.output_optimizer = None
        self.__ctx = ctx

        if self.init_deferred_flag is False:
            if self.__not_init_flag is False:
                if self.encoder_optimizer_f is not None:
                    self.encoder_optimizer = self.encoder_optimizer_f(
                        self.encoder.parameters(), 
                    )
                elif self.optimizer_f is not None:
                    self.encoder_optimizer = self.optimizer_f(
                        self.encoder.parameters(), 
                    )
                else:
                    self.encoder_optimizer = Adam(
                        self.encoder.parameters(), 
                        lr=self.__learning_rate,
                    )

                if self.decoder_optimizer_f is not None:
                    self.decoder_optimizer = self.decoder_optimizer_f(
                        self.decoder.parameters(), 
                    )
                elif self.optimizer_f is not None:
                    self.decoder_optimizer = self.optimizer_f(
                        self.decoder.parameters(), 
                    )
                else:
                    self.decoder_optimizer = Adam(
                        self.decoder.parameters(), 
                        lr=self.__learning_rate,
                    )

    def parameters(self):
        '''
        '''
        params_dict_list = [
            {
                "params": self.encoder.parameters(),
            },
            {
                "params": self.decoder.parameters(),
            }
        ]
        if self.output_nn is not None:
            params_dict_list.append(
                {
                    "params": self.output_nn.parameters()
                }
            )
        return params_dict_list

    def learn(self, iteratable_data):
        '''
        Learn the observed data points
        for vector representation of the input images.

        Args:
            iteratable_data:     is-a `IteratableData`.

        '''
        if isinstance(iteratable_data, IteratableData) is False:
            raise TypeError("The type of `iteratable_data` must be `IteratableData`.")

        self.__loss_list = []
        learning_rate = self.__learning_rate
        try:
            epoch = 0
            iter_n = 0
            for batch_observed_arr, batch_target_arr, test_batch_observed_arr, test_batch_target_arr in iteratable_data.generate_learned_samples():
                self.epoch = epoch
                self.batch_size = batch_observed_arr.shape[0]
                self.encoder_optimizer.zero_grad()
                self.decoder_optimizer.zero_grad()

                # rank-3
                pred_arr = self.inference(batch_observed_arr)
                loss = self.compute_loss(
                    pred_arr,
                    batch_target_arr
                )
                loss.backward()
                self.encoder_optimizer.step()
                self.decoder_optimizer.step()
                self.regularize()

                if (iter_n+1) % int(iteratable_data.iter_n / iteratable_data.epochs) == 0:
                    with torch.inference_mode():
                        # rank-3
                        test_pred_arr = self.inference(test_batch_observed_arr)
                        test_loss = self.compute_loss(
                            test_pred_arr,
                            test_batch_target_arr
                        )
                    _loss = loss.to('cpu').detach().numpy().copy()
                    _test_loss = test_loss.to('cpu').detach().numpy().copy()

                    self.__loss_list.append((_loss, _test_loss))

                    self.__logger.debug(
                        "Epochs: " + str(epoch + 1) + " Train loss: " + str(_loss) + " Test loss: " + str(_test_loss)
                    )
                    epoch += 1
                iter_n += 1

        except KeyboardInterrupt:
            self.__logger.debug("Interrupt.")

        self.__logger.debug("end. ")

    def inference(self, observed_arr):
        '''
        Inference the feature points to reconstruct the observed data points.

        Args:
            observed_arr:           rank-4 array like or sparse matrix as the observed data points.
                                    The shape is: (batch size, channel, height, width)

        Returns:
            `tensor` of inferenced feature points.
        '''
        return self(observed_arr)

    def compute_loss(self, pred_arr, labeled_arr):
        '''
        Compute loss.

        Args:
            pred_arr:       `tensor`.
            labeled_arr:    `tensor`.

        Returns:
            loss.
        '''
        return self.__computable_loss(pred_arr, labeled_arr)

    def extract_feature_points(self):
        '''
        Extract the activities in hidden layer and reset it.

        Returns:
            The `tensor` of array like or sparse matrix of feature points or virtual visible observed data points.
        '''
        return self.feature_points_arr

    def extract_learned_dict(self):
        '''
        Extract (pre-) learned parameters.

        Returns:
            `dict` of the parameters.
        '''
        params_arr_dict = {}

        params_dict = self.encoder.extract_learned_dict()
        for k in params_dict:
            params_arr_dict.setdefault(k, params_dict[k].data())

        params_dict = self.decoder.extract_learned_dict()
        for k in params_dict:
            params_arr_dict.setdefault(k, params_dict[k].data())

        return params_arr_dict

    def forward(self, x):
        '''
        Forward with Gluon API.

        Args:
            x:      `tensor` of observed data points.
        
        Returns:
            `tensor` of inferenced feature points.
        '''
        encoded_arr = self.encoder(x)
        self.feature_points_arr = encoded_arr
        if self.output_nn is None:
            decoded_arr = self.decoder(encoded_arr)
        else:
            inner_decoded_arr = self.output_nn(encoded_arr)
            decoded_arr = self.decoder(inner_decoded_arr)
        self.__pred_arr = decoded_arr
        return decoded_arr

    def regularize(self):
        '''
        Regularization.
        '''
        self.encoder.regularize()
        self.decoder.regularize()
        self.__tie_weights()

    def __tie_weights(self):
        if self.__tied_weights_flag is True:
            encoder_params_dict = self.encoder.extract_learned_dict()
            decoder_params_dict = self.decoder.extract_learned_dict()
            encoder_weight_keys_list = [key for key in encoder_params_dict.keys() if "hidden_units_list" in key and "weight" in key]
            decoder_weight_keys_list = [key for key in decoder_params_dict.keys() if "hidden_units_list" in key and "weight" in key]

            if len(encoder_weight_keys_list) != len(decoder_weight_keys_list):
                raise ValueError(
                    "The number of layers is invalid."
                )

            for i in range(len(self.encoder.hidden_units_list)):
                encoder_layer = i
                decoder_layer = len(self.encoder.hidden_units_list) - i - 1
                encoder_weight_keys, decoder_weight_keys = None, None
                for _encoder_weight_keys in encoder_weight_keys_list:
                    if "hidden_units_list." + str(encoder_layer) + ".weight" in _encoder_weight_keys:
                        encoder_weight_keys = _encoder_weight_keys
                        break

                for _decoder_weight_keys in decoder_weight_keys_list:
                    if "hidden_units_list." + str(decoder_layer) + ".weight" in _decoder_weight_keys:
                        decoder_weight_keys = _decoder_weight_keys
                        break

                if encoder_weight_keys is not None and decoder_weight_keys is not None:
                    try:
                        decoder_params_dict[decoder_weight_keys] = encoder_params_dict[encoder_weight_keys]
                    except AssertionError:
                        raise ValueError(
                            "The shapes of weight matrixs must be equivalents in encoder layer " + str(encoder_layer) + " and decoder layer " + str(decoder_layer)
                        )

            for k, params in decoder_params_dict.items():
                if k in decoder_weight_keys_list:
                    self.decoder.load_state_dict({k: params}, strict=False)

    def set_readonly(self, value):
        ''' setter '''
        raise TypeError("This property must be read-only.")

    def get_loss_arr(self):
        ''' getter for losses. '''
        return np.array(self.__loss_list)

    loss_arr = property(get_loss_arr, set_readonly)

    def get_init_deferred_flag(self):
        ''' getter for `bool` that means initialization in this class will be deferred or not.'''
        return self.__init_deferred_flag
    
    def set_init_deferred_flag(self, value):
        ''' setter for `bool` that means initialization in this class will be deferred or not. '''
        self.__init_deferred_flag = value

    init_deferred_flag = property(get_init_deferred_flag, set_init_deferred_flag)

    def get_batch_size(self):
        ''' getter for batch size.'''
        return self.__batch_size
    
    def set_batch_size(self, value):
        ''' setter for batch size.'''
        self.__batch_size = value
    
    batch_size = property(get_batch_size, set_batch_size)

    def get_computable_loss(self):
        ''' getter for `ComputableLoss`.'''
        return self.__computable_loss
    
    def set_computable_loss(self, value):
        ''' setter for `ComputableLoss`.'''
        self.__computable_loss = value
    
    computable_loss = property(get_computable_loss, set_computable_loss)



# -*- coding: utf-8 -*-
from accelbrainbase.iteratabledata._torch.ssda_iterator import SSDAIterator
import numpy as np
import torch
from abc import abstractmethod


class RotationIterator(SSDAIterator):
    '''
    Iterator that draws from image files and generates `mxnet.ndarray`.

    References:
        - Jing, L., & Tian, Y. (2020). Self-supervised visual feature learning with deep neural networks: A survey. IEEE Transactions on Pattern Analysis and Machine Intelligence.
        - Xu, J., Xiao, L., & López, A. M. (2019). Self-supervised domain adaptation for computer vision tasks. IEEE Access, 7, 156694-156706., p156698.
    '''

    def create_pretext_task_samples(self, target_domain_batch_arr):
        '''
        Create samples for pretext_task.

        Args:
            target_domain_batch_arr:    `nd.ndarray` of samples in target domain.
        
        Returns:
            Tuple data. The shape is ...
            - pretext_task samples.
            - pretext_task labels.
        '''
        _target_domain_batch_arr = target_domain_batch_arr.detach()
        angle_list = []
        for batch in range(target_domain_batch_arr.shape[0]):
            angle_key = np.random.randint(low=0, high=4)
            angle_arr = np.zeros(4)
            angle_arr[angle_key] = 1
            angle = 90 * angle_key
            _target_domain_batch_arr[batch] = self.__rotate(_target_domain_batch_arr[batch], angle)

            angle_list.append(angle_arr)

        angle_arr = torch.from_numpy(np.array(angle_list))
        angle_arr = angle_arr.to(pretext_arr.device)

        return _target_domain_batch_arr, angle_arr

    def __rotate(self, arr, angle=0):
        if angle == 0:
            return arr
        elif angle == 90:
            arr = arr.reshape((arr.shape[0], arr.shape[2], arr.shape[1]))
            return arr
        elif angle == 180:
            arr = torch.flip(arr, dims=[1])
            return arr
        elif angle == 270:
            arr = arr.reshape((arr.shape[0], arr.shape[2], arr.shape[1]))
            arr = torch.flip(arr, dims=[1])
            return arr



# -*- coding: utf-8 -*-
import numpy as np
import mxnet.ndarray as nd
import mxnet as mx
from mxnet import gluon
from mxnet import MXNetError

import warnings
from accelbrainbase.observabledata._mxnet.restricted_boltzmann_machines import RestrictedBoltzmannMachines
from accelbrainbase.iteratable_data import IteratableData
from accelbrainbase._mxnet._exception.init_deferred_error import InitDeferredError
from logging import getLogger


class DeepBoltzmannMachines(RestrictedBoltzmannMachines):
    '''
    Deep Boltzmann Machines(DBM).
    
    As is well known, DBM is composed of layers of RBMs 
    stacked on top of each other(Salakhutdinov, R., & Hinton, G. E. 2009). 
    This model is a structural expansion of Deep Belief Networks(DBN), 
    which is known as one of the earliest models of Deep Learning
    (Le Roux, N., & Bengio, Y. 2008). Like RBM, DBN places nodes in layers. 
    However, only the uppermost layer is composed of undirected edges, 
    and the other consists of directed edges.

    According to graph theory, the structure of RBM corresponds to 
    a complete bipartite graph which is a special kind of bipartite 
    graph where every node in the visible layer is connected to every 
    node in the hidden layer. Based on statistical mechanics and 
    thermodynamics(Ackley, D. H., Hinton, G. E., & Sejnowski, T. J. 1985), 
    the state of this structure can be reflected by the energy function.

    In relation to RBM, the Contrastive Divergence(CD) is a method for 
    approximation of the gradients of the log-likelihood(Hinton, G. E. 2002).
    This algorithm draws a distinction between a positive phase and a 
    negative phase. Conceptually, the positive phase is to the negative 
    phase what waking is to sleeping.

    The procedure of this method is similar to Markov Chain Monte Carlo method(MCMC).
    However, unlike MCMC, the visbile variables to be set first in visible layer is 
    not randomly initialized but the observed data points in training dataset are set 
    to the first visbile variables. And, like Gibbs sampler, drawing samples from hidden 
    variables and visible variables is repeated k times. Empirically (and surprisingly), 
    `k` is considered to be `1`.

    **Note** that this class does not support a *Hybrid* of imperative and symbolic 
    programming. Only `mxnet.ndarray` is supported.

    References:
        - Ackley, D. H., Hinton, G. E., & Sejnowski, T. J. (1985). A learning algorithm for Boltzmann machines. Cognitive science, 9(1), 147-169.
        - Hinton, G. E. (2002). Training products of experts by minimizing contrastive divergence. Neural computation, 14(8), 1771-1800.
        - Le Roux, N., & Bengio, Y. (2008). Representational power of restricted Boltzmann machines and deep belief networks. Neural computation, 20(6), 1631-1649.
        - Salakhutdinov, R., & Hinton, G. E. (2009). Deep boltzmann machines. InInternational conference on artificial intelligence and statistics (pp. 448-455).

    '''

    # `bool` that means initialization in this class will be deferred or not.
    __init_deferred_flag = False

    def __init__(
        self, 
        computable_loss,
        rbm_list,
        initializer=None,
        optimizer_name="SGD",
        learning_rate=0.005,
        learning_attenuate_rate=1.0,
        attenuate_epoch=50,
        ctx=mx.gpu(),
        **kwargs
    ):
        '''
        Init.
        
        Args:
            computable_loss:            is-a `ComputableLoss`.
            rbm_list:                   `list` of `RestrictedBoltzmannMachines`s.
            initializer:                is-a `mxnet.initializer` for parameters of model. If `None`, it is drawing from the Xavier distribution.
            optimizer_name:             `str` of name of optimizer.
            learning_rate:              `float` of learning rate.
            learning_attenuate_rate:    `float` of attenuate the `learning_rate` by a factor of this value every `attenuate_epoch`.
            attenuate_epoch:            `int` of attenuate the `learning_rate` by a factor of `learning_attenuate_rate` every `attenuate_epoch`.
            ctx:                        `mx.gpu()` or `mx.cpu()`.

        '''
        for rbm in rbm_list:
            if isinstance(rbm, RestrictedBoltzmannMachines) is False:
                raise TypeError("The type of values of `rbm_list` must be `RestrictedBoltzmannMachines`.")
        self.__rbm_list = rbm_list

        init_deferred_flag = self.init_deferred_flag
        self.init_deferred_flag = True
        super().__init__(
            computable_loss=computable_loss,
            initializer=initializer,
            optimizer_name=optimizer_name,
            learning_rate=learning_rate,
            learning_attenuate_rate=learning_attenuate_rate,
            attenuate_epoch=attenuate_epoch,
            ctx=ctx,
            **kwargs
        )
        self.init_deferred_flag = init_deferred_flag

        if initializer is None:
            self.initializer = mx.initializer.Xavier(
                rnd_type="gaussian", 
                factor_type="in", 
                magnitude=1
            )
        else:
            if isinstance(initializer, mx.initializer.Initializer) is False:
                raise TypeError("The type of `initializer` must be `mxnet.initializer.Initializer`.")

            self.initializer = initializer

        if self.init_deferred_flag is False:
            try:
                self.collect_params().initialize(self.initializer, force_reinit=True, ctx=ctx)
                self.trainer = gluon.Trainer(self.collect_params(), optimizer_name, {"learning_rate": learning_rate})
            except InitDeferredError:
                self.__logger.debug("The initialization should be deferred.")

        logger = getLogger("accelbrainbase")
        self.__logger = logger

    def collect_params(self, select=None):
        '''
        Overrided `collect_params` in `mxnet.gluon.HybridBlok`.
        '''
        params_dict = self.__rbm_list[0].collect_params(select)
        for i in range(1, len(self.__rbm_list)):
            params_dict.update(self.__rbm_list[i].collect_params(select))
        return params_dict

    def inference(self, observed_arr):
        '''
        Inference samples drawn by `IteratableData.generate_inferenced_samples()`.

        Args:
            observed_arr:   rank-2 Array like or sparse matrix as the observed data points.
                            The shape is: (batch size, feature points)

        Returns:
            `mxnet.ndarray` of inferenced feature points.
        '''
        self.batch_size = observed_arr.shape[0]
        for i in range(len(self.__rbm_list)):
            self.__rbm_list[i].batch_size = self.batch_size
            _ = self.__rbm_list[i].inference(observed_arr)
            observed_arr = self.__rbm_list[i].feature_points_arr

        return observed_arr

    def regularize(self):
        '''
        Regularization.
        '''
        for i in range(len(self.__rbm_list)):
            self.__rbm_list[i].regularize()

    def set_readonly(self, value):
        ''' setter '''
        raise TypeError("This is read-only.")

    def get_feature_points_arr(self):
        ''' getter for `mxnet.ndarray` of feature points.'''
        return self.__hidden_activity_arr

    feature_points_arr = property(get_feature_points_arr, set_readonly)

    def get_init_deferred_flag(self):
        ''' getter for `bool` that means initialization in this class will be deferred or not.'''
        return self.__init_deferred_flag
    
    def set_init_deferred_flag(self, value):
        ''' setter for `bool` that means initialization in this class will be deferred or not.'''
        self.__init_deferred_flag = value

    init_deferred_flag = property(get_init_deferred_flag, set_init_deferred_flag)

    def get_rbm_list(self):
        ''' getter '''
        return self.__rbm_list
    
    def set_rbm_list(self, value):
        ''' setter '''
        self.__rbm_list = value
    
    rbm_list = property(get_rbm_list, set_rbm_list)



# -*- coding: utf-8 -*-
from accelbrainbase.computable_loss import ComputableLoss

import torch
import torch.nn as nn

import numpy as np


class AdversarialSSDALoss(nn.modules.loss._Loss, ComputableLoss):
    '''
    Loss function of Self-supervised domain adaptation.

    References:
        - Jing, L., & Tian, Y. (2020). Self-supervised visual feature learning with deep neural networks: A survey. IEEE Transactions on Pattern Analysis and Machine Intelligence.
        - Xu, J., Xiao, L., & López, A. M. (2019). Self-supervised domain adaptation for computer vision tasks. IEEE Access, 7, 156694-156706., p156698.
    '''

    def __init__(
        self, 
        weight=1.0,
        classification_loss_f=None,
        pretext_loss_f=None,
        classification_weight=None, 
        pretext_weight=None, 
        adversarial_weight=None,
        adversarial_lambda=0.5,
        batch_axis=1, 
        label_smoothing=0.0,
    ):
        '''
        Init.

        Args:
            axis:                   The axis to sum over when computing softmax and entropy.
            sparse_label:           Whether label is an integer array instead of probability distribution.
            from_logits:            Whether input is a log probability (usually from log_softmax) instead of unnormalized numbers.
            log_softmax_flag:       Use `F.log_softmax` when `from_logits` is `False`.
                                    If this value is `False`, this class will use not `F.log_softmax` but `F.softmax`.

            weight:                 Global scalar weight for total loss.
            classification_weight:  Global scalar weight for classification loss. If `None`, this value will be equivalent to `weight`.
            pretext_weight:         Global scalar weight for pretext loss. If `None`, this value will be equivalent to `weight`.
            adversarial_weight:     Global scalar weight for adversarial loss. If `None`, this value will be equivalent to `weight`.
            adversarial_lambda:     Trade-off parameter for adversarial loss. $L_{d} = (1 - `adversarial_lambda`) * {target_posterior_arr} + `adversarial_lambda` * {source_posterior_arr}.
            grad_clip_threshold:    Threshold of the gradient clipping.
            batch_axis:             The axisthat represents mini-batch.
            **kwargs:               Any parmaeter.
        '''
        super(AdversarialSSDALoss, self).__init__()

        if classification_loss_f is not None:
            self.__classification_loss_f = classification_loss_f
        else:
            self.__classification_loss_f = torch.nn.CrossEntropyLoss(
                size_average=None, 
                ignore_index=-100, 
                reduce=None, 
                reduction='mean', 
                label_smoothing=label_smoothing
            )

        if pretext_loss_f is not None:
            self.__pretext_loss_f = pretext_loss_f
        else:
            self.__pretext_loss_f = torch.nn.CrossEntropyLoss(
                size_average=None, 
                ignore_index=-100, 
                reduce=None, 
                reduction='mean', 
                label_smoothing=label_smoothing
            )

        self.__weight = weight
        self.__classification_weight = classification_weight
        self.__pretext_weight = pretext_weight
        self.__adversarial_weight = adversarial_weight
        self.__adversarial_lambda = adversarial_lambda
        self.__batch_axis = batch_axis

    def compute(
        self, 
        pretext_pred_arr, 
        pred_arr, 
        pretext_label_arr, 
        label_arr, 
        source_posterior_arr,
        target_posterior_arr,
    ):
        '''
        Compute loss.

        Args:
            pretext_pred_arr:       `mxnet.ndarray` or `mxnet.symbol` of decoded feature points.
            pred_arr:               `mxnet.ndarray` or `mxnet.symbol` of inferenced labeled feature points.
            pretext_label_arr:      `mxnet.ndarray` or `mxnet.symbol` of observed data points.
            label_arr:              `mxnet.ndarray` or `mxnet.symbol` of label data.
            source_posterior_arr:     `mxnet.ndarray` or `mxnet.symbol` of encoded data points in source domain.
            target_posterior_arr:     `mxnet.ndarray` or `mxnet.symbol` of encoded data poitns in target domain.

        Returns:
            Tensor of losses.
        '''
        return self(
            pretext_pred_arr, 
            pred_arr, 
            pretext_label_arr, 
            label_arr, 
            source_posterior_arr,
            target_posterior_arr,
        )

    def forward(
        self, 
        pretext_pred_arr, 
        pred_arr, 
        pretext_label_arr, 
        label_arr, 
        source_posterior_arr,
        target_posterior_arr,
    ):
        '''
        Forward propagation, computing losses.

        Args:
            F:                      `mxnet.ndarray` or `mxnet.symbol`.
            pretext_pred_arr:       `mxnet.ndarray` or `mxnet.symbol` of decoded feature points.
            pred_arr:               `mxnet.ndarray` or `mxnet.symbol` of inferenced labeled feature points.
            pretext_label_arr:      `mxnet.ndarray` or `mxnet.symbol` of observed data points.
            label_arr:              `mxnet.ndarray` or `mxnet.symbol` of label data.
            source_posterior_arr:     `mxnet.ndarray` or `mxnet.symbol` of encoded data points in source domain.
            target_posterior_arr:     `mxnet.ndarray` or `mxnet.symbol` of encoded data poitns in target domain.

            sample_weight:          element-wise weighting tensor. 
                                    Must be broadcastable to the same shape as label. 
                                    For example, if label has shape (64, 10) and you want to weigh 
                                    each sample in the batch separately, sample_weight should have shape (64, 1).

        Returns:
            `mxnet.ndarray` or `mxnet.symbol` of loss.
        '''
        classification_loss = self.__classification_loss_f(
            pred_arr,
            label_arr
        )
        if self.__classification_weight is not None:
            classification_loss = classification_loss * self.__classification_weight

        classification_loss = classification_loss * self.__weight

        pretext_label_arr = pretext_label_arr.reshape_as(pretext_pred_arr)

        pretext_loss = self.__pretext_loss_f(
            pretext_pred_arr,
            pretext_label_arr
        )
        if self.__pretext_weight is not None:
            pretext_loss = pretext_loss * self.__pretext_weight

        pretext_loss = pretext_loss * self.__weight

        adversarial_loss_arr = - ((1 - self.__adversarial_lambda) * torch.log(target_posterior_arr) + self.__adversarial_lambda * torch.log(source_posterior_arr))
        adversarial_loss_arr = torch.reshape(adversarial_loss_arr, (adversarial_loss_arr.shape[0], -1))
        adversarial_loss = torch.mean(adversarial_loss_arr)

        if self.__adversarial_weight is not None:
            adversarial_loss = adversarial_loss * self.__adversarial_weight

        adversarial_loss = adversarial_loss * self.__weight

        total_loss = classification_loss + pretext_loss + adversarial_loss
        return total_loss, classification_loss, pretext_loss, adversarial_loss



# -*- coding: utf-8 -*-
from accelbrainbase.observable_data import ObservableData
from abc import abstractproperty
import torch
from torch import nn


class FunctionApproximator(ObservableData):
    '''
    The function approximator for the Deep Q-Learning.

    The convolutional neural networks(CNNs) are hierarchical models 
    whose convolutional layers alternate with subsampling layers, 
    reminiscent of simple and complex cells in the primary visual cortex.
    
    Mainly, this class demonstrates that a CNNs can solve generalisation problems to learn 
    successful control policies from observed data points in complex 
    Reinforcement Learning environments. The network is trained with a variant of 
    the Q-learning algorithm, with stochastic gradient descent to update the weights.

    But there is no need for the function approximator to be a CNNs.
    We probide this interface that implements various models as 
    function approximations, not limited to CNNs.

    References:
        - Dumoulin, V., & V,kisin, F. (2016). A guide to convolution arithmetic for deep learning. arXiv preprint arXiv:1603.07285.
        - Masci, J., Meier, U., Cireşan, D., & Schmidhuber, J. (2011, June). Stacked convolutional auto-encoders for hierarchical feature extraction. In International Conference on Artificial Neural Networks (pp. 52-59). Springer, Berlin, Heidelberg.
        - Mnih, V., Kavukcuoglu, K., Silver, D., Graves, A., Antonoglou, I., Wierstra, D., & Riedmiller, M. (2013). Playing atari with deep reinforcement learning. arXiv preprint arXiv:1312.5602.

    '''

    def learn(self, iteratable_data):
        '''
        Learn the observed data points
        for vector representation of the input images.

        Args:
            iteratable_data:     is-a `IteratableData`.

        '''
        self.model.learn(iteratable_data)

    def inference(self, observed_arr):
        '''
        Inference samples drawn by `IteratableData.generate_inferenced_samples()`.

        Args:
            observed_arr:   rank-2 Array like or sparse matrix as the observed data points.
                            The shape is: (batch size, feature points)

        Returns:
            `mxnet.ndarray` of inferenced feature points.
        '''
        return self.model.inference(observed_arr)

    # `bool` that means initialization in this class will be deferred or not.
    __init_deferred_flag = False

    def get_init_deferred_flag(self):
        ''' getter for `bool` that means initialization in this class will be deferred or not. '''
        return self.__init_deferred_flag
    
    def set_init_deferred_flag(self, value):
        ''' setter for `bool` that means initialization in this class will be deferred or not. '''
        self.__init_deferred_flag = value

    init_deferred_flag = property(get_init_deferred_flag, set_init_deferred_flag)



# -*- coding: utf-8 -*-
import numpy as np
import mxnet as mx
import mxnet.ndarray as nd
import pandas as pd
from logging import getLogger
import os

from accelbrainbase.extractabledata.unlabeled_csv_extractor import UnlabeledCSVExtractor
from accelbrainbase.iteratabledata.unlabeled_image_iterator import UnlabeledImageIterator as _UnlabeledImageIterator
from accelbrainbase.noiseable_data import NoiseableData


class UnlabeledTHotTXTIterator(_UnlabeledImageIterator):
    '''
    Iterator that draws from CSV files and generates `mxnet.ndarray` of unlabeled samples.
    '''

    def get_token_list(self):
        return self.__token_list
    
    def set_token_list(self, value):
        self.__token_list = value
    
    token_list = property(get_token_list, set_token_list)

    def get_token_arr(self):
        return np.array(self.__token_list)
    
    def set_token_arr(self, value):
        raise TypeError("This property must be read-only.")
    
    token_arr = property(get_token_arr, set_token_arr)

    def get_pre_txt_arr(self):
        return self.__pre_txt_arr
    
    def set_pre_txt_arr(self, value):
        self.__pre_txt_arr = value

    pre_txt_arr = property(get_pre_txt_arr, set_pre_txt_arr)

    def __init__(
        self,
        train_txt_path_list,
        test_txt_path_list=None,
        epochs=300,
        batch_size=20,
        seq_len=10,
        norm_mode="z_score",
        scale=1.0,
        noiseable_data=None,
        dataset_size=None,
        ctx=mx.gpu()
    ):
        '''
        Init.

        Args:
            image_extractor:                is-a `ImageExtractor`.
            train_txt_path_list:            `list` of `str` of path to CSV file in training.
            test_txt_path_list:             `list` of `str` of path to CSV file in test.

            epochs:                         `int` of epochs of Mini-batch.
            bath_size:                      `int` of batch size of Mini-batch.
            seq_len:                        `int` of the length of series. 
            norm_mode:                      How to normalize pixel values of images.
                                            - `z_score`: Z-Score normalization.
                                            - `min_max`: Min-max normalization.
                                            - others : This class will not normalize the data.

            scale:                          `float` of scaling factor for data.
            noiseable_data:                 is-a `NoiseableData` for Denoising Auto-Encoders.
            dataset_size:                   `int` of the dataset size.
                                            If `None`, this class will consider `dataset_size` as 
                                            the number of text in training data.
        '''
        if noiseable_data is not None and isinstance(noiseable_data, NoiseableData) is False:
            raise TypeError("The type of `noiseable_data` must be `NoiseableData`.")

        logger = getLogger("accelbrainbase")
        self.__logger = logger

        train_txt = ""
        for train_txt_path in train_txt_path_list:
            with open(train_txt_path) as f:
                train_txt += f.read()

        if dataset_size is None:
            dataset_size = len(train_txt)

        iter_n = int(epochs * max(dataset_size / batch_size, 1))

        train_token_list = list(train_txt)
        self.__token_list = list(set(train_token_list))

        self.__train_txt_path_list = train_txt_path_list
        if test_txt_path_list is not None:
            self.__test_txt_path_list = test_txt_path_list
        else:
            self.__test_txt_path_list = train_txt_path_list

        self.iter_n = iter_n
        self.epochs = epochs
        self.batch_size = batch_size
        self.seq_len = seq_len
        self.norm_mode = norm_mode
        self.scale = scale
        self.__noiseable_data = noiseable_data

        self.__ctx = ctx

        self.__pre_txt_arr = None

    def generate_learned_samples(self):
        '''
        Draw and generate data.

        Returns:
            `Tuple` data. The shape is ...
            - `mxnet.ndarray` of observed data points in training.
            - `mxnet.ndarray` of supervised data in training.
            - `mxnet.ndarray` of observed data points in test.
            - `mxnet.ndarray` of supervised data in test.
        '''
        for _ in range(self.iter_n):
            training_batch_arr = np.zeros(
                (
                    self.batch_size,
                    1,
                    self.seq_len,
                    len(self.__token_list)
                ),
            )
            test_batch_arr = np.zeros(
                (
                    self.batch_size,
                    1,
                    self.seq_len,
                    len(self.__token_list)
                ),
            )

            for i in range(self.batch_size):
                file_key = np.random.randint(low=0, high=len(self.__train_txt_path_list))
                with open(self.__train_txt_path_list[file_key]) as f:
                    train_txt = f.read()

                test_file_key = np.random.randint(low=0, high=len(self.__test_txt_path_list))
                with open(self.__test_txt_path_list[test_file_key]) as f:
                    test_txt = f.read()
                
                if self.__pre_txt_arr is None:
                    start_row = np.random.randint(low=0, high=len(train_txt) - self.seq_len)
                    test_start_row = np.random.randint(low=0, high=len(test_txt) - self.seq_len)

                    train_txt = train_txt[start_row:start_row+self.seq_len]
                    test_txt = test_txt[test_start_row:test_start_row+self.seq_len]
                else:
                    if train_txt.index(self.__pre_txt_arr[i]) + (self.seq_len * 2) < len(train_txt):
                        start_row = train_txt.index(self.__pre_txt_arr[i]) + self.seq_len
                    else:
                        start_row = np.random.randint(low=0, high=len(train_txt) - self.seq_len)
                    train_txt = train_txt[start_row:start_row+self.seq_len]

                    test_start_row = np.random.randint(low=0, high=len(test_txt) - self.seq_len)
                    test_txt = test_txt[test_start_row:test_start_row+self.seq_len]


                for seq in range(self.seq_len):
                    training_batch_arr[i, 0, seq, self.__token_list.index(train_txt[seq])] = 1.0
                    test_batch_arr[i, 0, seq, self.__token_list.index(test_txt[seq])] = 1.0

            training_batch_arr = nd.ndarray.array(training_batch_arr, ctx=self.__ctx)
            test_batch_arr = nd.ndarray.array(test_batch_arr, ctx=self.__ctx)

            training_batch_arr = self.pre_normalize(training_batch_arr)
            test_batch_arr = self.pre_normalize(test_batch_arr)

            if self.__noiseable_data is not None:
                training_batch_arr = self.__noiseable_data.noise(training_batch_arr)

            yield training_batch_arr, training_batch_arr, test_batch_arr, test_batch_arr

    def generate_inferenced_samples(self):
        '''
        Draw and generate data.
        The targets will be drawn from all image file sorted in ascending order by file name.

        Returns:
            `Tuple` data. The shape is ...
            - `None`.
            - `None`.
            - `mxnet.ndarray` of observed data points in test.
            - file path.
        '''
        for test_file_key in range(len(self.__test_txt_path_list)):
            global_seq = 0
            with open(self.__test_txt_path_list[test_file_key]) as f:
                test_txt = f.read()

            _test_batch_arr = np.zeros(
                (
                    self.batch_size,
                    1,
                    self.seq_len,
                    len(self.__token_list)
                ),
            )
            batch = 0
            for start_row in range(len(test_txt) - (self.seq_len)):
                for seq in range(self.seq_len):
                    _test_batch_arr[
                        batch, 
                        0, 
                        seq, 
                        self.__token_list.index(test_txt[start_row+seq])
                    ] = 1.0
                batch += 1
                if batch == self.batch_size:
                    test_batch_arr = nd.ndarray.array(_test_batch_arr, ctx=self.__ctx)
                    test_batch_arr = self.pre_normalize(test_batch_arr)
                    yield None, None, test_batch_arr, None
                    batch = 0
                    _test_batch_arr = np.zeros(
                        (
                            self.batch_size,
                            1,
                            self.seq_len,
                            len(self.__token_list)
                        ),
                    )



# -*- coding: utf-8 -*-
from accelbrainbase.samplabledata.policy_sampler import PolicySampler
from accelbrainbase.iteratabledata.unlabeled_image_iterator import UnlabeledImageIterator

from accelbrainbase.computableloss._mxnet.l2_norm_loss import L2NormLoss
import mxnet.ndarray as nd
import mxnet as mx


class UnlabeledSimilarImagePolicy(PolicySampler):
    '''
    Policy sampler for the Deep Q-learning to evaluate the value of the 
    "action" of selecting the image with the highest similarity based on 
    the "state" of observing an image.

    The state-action value is proportional to the similarity between the previously 
    observed image and the currently selected image.

    This class calculates the image similarity by mean squared error of images.
    '''

    # is-a `UnlabeledImageIterator`.
    __unlabeled_image_iterator = None

    def get_unlabeled_image_iterator(self):
        ''' getter for `UnlabeledImageIterator`. '''
        if isinstance(self.__unlabeled_image_iterator, UnlabeledImageIterator) is False:
            raise TypeError("The type of `__unlabeled_image_iterator` must be `UnlabeledImageIterator`.")
        return self.__unlabeled_image_iterator
    
    def set_unlabeled_image_iterator(self, value):
        ''' setter for `UnlabeledImageIterator`.'''
        if isinstance(value, UnlabeledImageIterator) is False:
            raise TypeError("The type of `__unlabeled_image_iterator` must be `UnlabeledImageIterator`.")
        self.__unlabeled_image_iterator = value

    unlabeled_image_iterator = property(get_unlabeled_image_iterator, set_unlabeled_image_iterator)

    def draw(self):
        '''
        Draw samples from distribtions.
        
        Returns:
            `Tuple` of `mx.nd.array`s.
        '''
        observed_arr, label_arr = None, None
        for arr_tuple in self.unlabeled_image_iterator.generate_learned_samples():
            if observed_arr is None:
                observed_arr = nd.expand_dims(arr_tuple[0], axis=1)
            else:
                observed_arr = nd.concat(
                    observed_arr, 
                    nd.expand_dims(arr_tuple[0], axis=1), 
                    dim=1
                )

            if observed_arr.shape[1] >= self.possible_n:
                break

        return observed_arr, None

    def observe_state(self, state_arr, meta_data_arr):
        '''
        Observe states of agents in last epoch.

        Args:
            state_arr:      Tensor of state.
            meta_data_arr:  meta data of the state.
        '''
        self.__state_arr = state_arr
        self.__state_meta_data_arr = meta_data_arr

    def observe_reward_value(
        self, 
        state_arr, 
        action_arr,
        meta_data_arr=None,
    ):
        '''
        Compute the reward value.
        
        Args:
            state_arr:              Tensor of state.
            action_arr:             Tensor of action.
            meta_data_arr:          Meta data of actions.

        Returns:
            Reward value.
        '''
        if state_arr is not None:
            mse_arr = nd.mean(
                nd.square(
                    nd.flatten(state_arr),
                    nd.flatten(action_arr)
                ),
                axis=0, 
                exclude=True
            )
            reward_value_arr = 1 / mse_arr
            reward_value_arr = nd.expand_dims(reward_value_arr, axis=1)
        else:
            reward_value_arr = nd.zeros((
                action_arr.shape[0],
                1
            ), ctx=action_arr.context)

        return reward_value_arr

    def update_state(
        self, 
        action_arr, 
        meta_data_arr=None
    ):
        '''
        Update state.
        
        This method can be overrided for concreate usecases.

        Args:
            action_arr:     action in `self.t`.
            meta_data_arr:  meta data of the action.
        
        Returns:
            Tuple data.
            - state in `self.t+1`.
            - meta data of the state.
        '''
        return action_arr, meta_data_arr

    def check_the_end_flag(self, state_arr, meta_data_arr=None):
        '''
        Check the end flag.
        
        If this return value is `True`, the learning is end.

        As a rule, the learning can not be stopped.
        This method should be overrided for concreate usecases.

        Args:
            state_arr:      state in `self.t`.
            meta_data_arr:  meta data of the state.

        Returns:
            bool
        '''
        # As a rule, the learning can not be stopped.
        return False



# -*- coding: utf-8 -*-
from accelbrainbase.samplabledata.true_sampler import TrueSampler
import mxnet.ndarray as nd
import mxnet as mx


class NormalTrueSampler(TrueSampler):
    '''
    The class to draw fake samples from Gaussian distributions.

    References:
        - Goodfellow, I., Pouget-Abadie, J., Mirza, M., Xu, B., Warde-Farley, D., Ozair, S., ... & Bengio, Y. (2014). Generative adversarial nets. In Advances in neural information processing systems (pp. 2672-2680).
        - Makhzani, A., Shlens, J., Jaitly, N., Goodfellow, I., & Frey, B. (2015). Adversarial autoencoders. arXiv preprint arXiv:1511.05644.
    '''

    def __init__(
        self, 
        loc=0.0,
        scale=1.0,
        batch_size=40,
        seq_len=0,
        channel=3,
        height=96,
        width=96,
        shape_tuple=None,
        ctx=mx.gpu()
    ):
        '''
        Init.

        Args:
            loc:            `float` of mean(centre) of the distribution.
            scale:          `float` of standard deviation(spread or width) of the distribution.

            batch_size:     `int` of batch size.
            seq_len:        `int` of the length of series.
                            If this value is `0`, the rank of matrix generated is `4`. 
                            The shape is: (`batch_size`, `channel`, `height`, `width`).
                            If this value is more than `0`, the rank of matrix generated is `5`.
                            The shape is: (`batch_size`, `seq_len`, `channel`, `height`, `width`).

            channel:        `int` of channel.
            height:         `int` of image height.
            width:          `int` of image width.
            shape_tuple:    `tuple` of shape of samples.
                            If this value is not `None`,  `batch_size`, `seq_len`, `channel`, `width` and `height` will be ignored.

            ctx:            `mx.gpu` or `mx.cpu`.
        '''
        import warnings
        warnings.warn("This class will be removed in future version. Use `accelbrainbase.samplabledata.truesampler._mxnet.normal_true_sampler.NormalTrueSampler`.", FutureWarning)

        self.__loc = loc
        self.__scale = scale
        self.__batch_size = batch_size
        self.__seq_len = seq_len
        self.__channel = channel
        self.__height = height
        self.__width = width
        self.__shape_tuple = shape_tuple
        self.__ctx = ctx

    def draw(self):
        '''
        Draw samples from distribtions.
        
        Returns:
            `Tuple` of `mx.nd.array`s.
        '''
        if self.__shape_tuple is None:
            if self.__seq_len > 0:
                shape_tuple = (
                    self.__batch_size,
                    self.__seq_len, 
                    self.__channel,
                    self.__height,
                    self.__width
                )
            else:
                shape_tuple = (
                    self.__batch_size,
                    self.__channel, 
                    self.__height,
                    self.__width
                )
        else:
            shape_tuple = self.__shape_tuple

        observed_arr = nd.random.normal(
            loc=self.__loc, 
            scale=self.__scale, 
            shape=shape_tuple,
            ctx=self.__ctx
        )
        return observed_arr



# -*- coding: utf-8 -*-
from accelbrainbase.computable_loss import ComputableLoss
import torch
import torch.nn as nn


class L2NormLoss(nn.modules.loss._Loss, ComputableLoss):
    '''
    Loss function that computes L2 norm.
    '''

    def __init__(self, weight=1.0, batch_axis=1):
        '''
        Init.

        Args:
            weight:         Global scalar weight for loss.
            batch_axis:     The axisthat represents mini-batch.
        '''
        self.__weight = weight
        self.__batch_axis = batch_axis
        super(L2NormLoss, self).__init__()

    def compute(self, pred_arr, real_arr):
        '''
        Compute loss.

        Args:
            pred_arr:       Inferenced results.
            real_arr:       Real results.
        
        Returns:
            Tensor of losses.
        '''
        return self(real_arr, pred_arr)

    def forward(self, orign_arr, dest_arr):
        '''
        Forward propagation, computing L2 norm.

        Args:
            orign_arr:   `tensor` of origins.
            dest_arr:    `tensor` of destinations.
        
        Returns:
            `mxnet.ndarray` or `mxnet.symbol` of loss.
        '''
        dest_arr = dest_arr.reshape_as(orign_arr)
        loss = torch.sqrt(torch.mean(torch.square(orign_arr - dest_arr), dim=self.__batch_axis, keepdim=True))
        loss = loss * self.__weight
        return torch.mean(loss)



# -*- coding: utf-8 -*-
from accelbrainbase.noiseabledata.gauss_noise import GaussNoise as _GaussNoise
import torch


class GaussNoise(_GaussNoise):
    '''
    Gauss noise function.
    '''

    def __init__(self, mu=0.0, sigma=1.0):
        '''
        Init.

        Args:
            mu:     Mean of the Gauss distribution.
            sigma:  Standard deviation of the Gauss distribution.
        '''
        self.__mu = mu
        self.__sigma = sigma

    def noise(self, arr, F=torch):
        '''
        Noise.

        Args:
            arr:    `mx.nd.array` or `mx.sym.array`.
            F:      `mx.ndarray` or `mx.symbol`.
        
        Returns:
            `mx.nd.array` or `mx.sym.array`.
        '''
        noise_arr = torch.normal(
            mean=self.__mu, 
            std=self.__sigma,
            size=arr.shape
        )
        noise_arr = noise_arr.to(arr.device)
        return arr + noise_arr
